# Copyright 2024 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import logging

import pmb.config
import pmb.helpers.git


def pull() -> None:
    failed = [repo for repo in pmb.config.git_repos if pmb.helpers.git.pull(repo) < 0]

    if not failed:
        return

    logging.info("---")
    logging.info("WARNING: failed to update: " + ", ".join(failed))
    logging.info("")
    logging.info("'pmbootstrap pull' will only update the repositories, if:")
    logging.info("* they are on an officially supported branch (e.g. main)")
    logging.info("* the history is not conflicting (fast-forward is possible)")
    logging.info("* the git workdirs are clean")
    logging.info("You have changed mentioned repositories, so they don't meet")
    logging.info("these conditions anymore.")
    logging.info("")
    logging.info("Fix and try again:")
    for name_repo in failed:
        logging.info(f"* {pmb.helpers.git.get_path(name_repo)}")
    logging.info("---")



# Copyright 2024 Caleb Connolly
# SPDX-License-Identifier: GPL-3.0-or-later

import enum
import inspect
import multiprocessing
import os
from copy import deepcopy
from pathlib import Path
from typing import Any, ClassVar, TypedDict


class Mirrors(TypedDict):
    alpine_custom: str
    alpine: str
    pmaports_custom: str
    pmaports: str
    systemd_custom: str
    systemd: str


class SystemdConfig(enum.Enum):
    DEFAULT = "default"
    ALWAYS = "always"
    NEVER = "never"

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def choices() -> list[str]:
        return [e.value for e in SystemdConfig]


class AutoZapConfig(enum.Enum):
    NO = "no"
    YES = "yes"
    SILENTLY = "silently"

    def __str__(self) -> str:
        return self.value

    def enabled(self) -> bool:
        return self != AutoZapConfig.NO

    def noisy(self) -> bool:
        return self == AutoZapConfig.YES


class Config:
    # This is a class variable that gets treated as an instance variable. It's wrong, but since we
    # only ever have one config (for now?) it doesn't cause any issues. Would be good to fix though.
    aports: list[Path] = [  # noqa: RUF012
        Path(os.path.expanduser("~") + "/.local/var/pmbootstrap/cache_git/pmaports")
    ]
    boot_size: int = 256
    build_default_device_arch: bool = False
    build_pkgs_on_install: bool = True
    ccache_size: str = "5G"  # yeahhhh this one has a suffix
    device: str = "qemu-amd64"
    extra_packages: str = "none"
    extra_space: int = 0
    hostname: str = ""
    is_default_channel: bool = True
    jobs: int = multiprocessing.cpu_count()
    kernel: str = "stable"
    keymap: str = ""
    locale: str = "en_US.UTF-8"
    mirrors: ClassVar[Mirrors] = {
        "alpine_custom": "none",
        "alpine": "http://dl-cdn.alpinelinux.org/alpine/",
        "pmaports_custom": "none",
        "pmaports": "http://mirror.postmarketos.org/postmarketos/",
        "systemd_custom": "none",
        "systemd": "http://mirror.postmarketos.org/postmarketos/extra-repos/systemd/",
    }
    qemu_redir_stdio: bool = False
    ssh_key_glob: str = "~/.ssh/*.pub"
    ssh_keys: bool = False
    sudo_timer: bool = False
    systemd: SystemdConfig = SystemdConfig.DEFAULT
    timezone: str = "GMT"
    ui: str = "console"
    ui_extras: bool = False
    user: str = "user"
    work: Path = Path(os.path.expanduser("~") + "/.local/var/pmbootstrap")
    # automatically zap chroots that are for the wrong channel
    auto_zap_misconfigured_chroots: AutoZapConfig = AutoZapConfig.NO

    providers: ClassVar[dict[str, str]] = {}

    def __init__(self) -> None:
        # Make sure we aren't modifying the class defaults
        for key in inspect.get_annotations(Config):
            setattr(self, key, deepcopy(Config.get_default(key)))

    @staticmethod
    def keys() -> list[str]:
        keys = list(inspect.get_annotations(Config).keys())
        keys.remove("mirrors")
        keys += [f"mirrors.{k}" for k in inspect.get_annotations(Mirrors)]
        return sorted(keys)

    @staticmethod
    def get_default(dotted_key: str) -> Any:
        """
        Get the default value for a config option, supporting
        nested dictionaries (e.g. "mirrors.alpine").
        """
        keys = dotted_key.split(".")
        if len(keys) == 1:
            return getattr(Config, keys[0])
        elif len(keys) == 2:
            return getattr(Config, keys[0])[keys[1]]
        else:
            raise ValueError(f"Invalid dotted key: {dotted_key}")

    def __setattr__(self, key: str, value: Any) -> None:
        """
        Allow for setattr() to be used with a dotted key
        to set nested dictionaries (e.g. "mirrors.alpine").
        """
        keys = key.split(".")
        if len(keys) == 1:
            type_ = type(getattr(Config, key))
            try:
                if type_ is bool and isinstance(value, str):
                    if value.lower() in ["true", "false"]:
                        super().__setattr__(key, value.lower() == "true")
                    else:
                        raise ValueError()
                else:
                    super().__setattr__(key, type_(value))
            except ValueError as exception:
                msg = f"Invalid value for '{key}': '{value}' "
                if issubclass(type_, enum.Enum):
                    valid = [x.value for x in type_]
                    msg += f"(valid values: {', '.join(valid)})"
                else:
                    msg += f"(expected {type_}, got {type(value)})"
                raise ValueError(msg) from exception
        elif len(keys) == 2:
            super().__getattribute__(keys[0])[keys[1]] = value
        else:
            raise ValueError(f"Invalid dotted key: {key}")

    def __getattribute__(self, key: str) -> Any:
        """
        Allow for getattr() to be used with a dotted key
        to get nested dictionaries (e.g. "mirrors.alpine").
        """
        keys = key.split(".")
        if len(keys) == 1:
            return super().__getattribute__(key)
        elif len(keys) == 2:
            return super().__getattribute__(keys[0])[keys[1]]
        else:
            raise ValueError(f"Invalid dotted key: {key}")



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
from pathlib import Path

import pmb.chroot.binfmt
import pmb.config
import pmb.helpers.mount
import pmb.helpers.run
import pmb.install.losetup
from pmb.core import Chroot
from pmb.core.chroot import ChrootType
from pmb.core.context import get_context
from pmb.core.pkgrepo import pkgrepo_default_path
from pmb.helpers import logging


def mount_chroot_image(chroot: Chroot) -> None:
    """
    Mount an IMAGE type chroot, to modify an existing rootfs image. This
    doesn't support split images yet!
    """
    # Make sure everything is nicely unmounted just to be super safe
    # this is definitely overkill
    pmb.chroot.shutdown()
    pmb.install.losetup.detach_all()

    chroot_native = Chroot.native()
    pmb.chroot.init(chroot_native)

    loopdev = pmb.install.losetup.mount(
        Path("/") / Path(chroot.name).relative_to(chroot_native.path)
    )
    pmb.helpers.mount.bind_file(loopdev, chroot_native / "dev/install")
    # Set up device mapper bits
    pmb.helpers.run.root(["kpartx", "-u", loopdev])
    chroot.path.mkdir(exist_ok=True)
    loopdev_basename = os.path.basename(loopdev)
    # # The name of the IMAGE chroot is the path to the rootfs image
    if Path(f"/dev/mapper/{loopdev_basename}p2").exists():
        pmb.helpers.run.root(["mount", f"/dev/mapper/{loopdev_basename}p2", chroot.path])
        pmb.helpers.run.root(["mount", f"/dev/mapper/{loopdev_basename}p1", chroot.path / "boot"])
    else:
        pmb.helpers.run.root(["mount", f"/dev/{loopdev_basename}", chroot.path])

    pmb.config.workdir.chroot_save_init(chroot)

    logging.info(f"({chroot}) mounted {chroot.name}")


def create_device_nodes(chroot: Chroot) -> None:
    """Create device nodes for null, zero, full, random, urandom in the chroot."""
    try:
        # Create all device nodes as specified in the config
        for dev in pmb.config.chroot_device_nodes:
            path = chroot / "dev" / str(dev[4])
            if not path.exists():
                pmb.helpers.run.root(
                    [
                        "mknod",
                        path,  # name
                        str(dev[1]),  # type
                        str(dev[2]),  # major
                        str(dev[3]),  # minor
                    ]
                )
                # chmod needs to be split from mknod to accommodate
                # for FreeBSD mknod not including -m
                pmb.helpers.run.root(
                    [
                        "chmod",
                        str(dev[0]),  # permissions
                        path,  # name
                    ]
                )

        # Verify major and minor numbers of created nodes
        for dev in pmb.config.chroot_device_nodes:
            path = chroot / "dev" / str(dev[4])
            stat_result = path.stat()
            rdev = stat_result.st_rdev
            assert os.major(rdev) == dev[2], f"Wrong major in {path}"
            assert os.minor(rdev) == dev[3], f"Wrong minor in {path}"

        # Verify /dev/zero reading and writing
        path = chroot / "dev/zero"
        with open(path, "r+b", 0) as handle:
            assert handle.write(bytes([0xFF])), f"Write failed for {path}"
            assert handle.read(1) == bytes([0x00]), f"Read failed for {path}"

    # On failure: Show filesystem-related error
    except Exception as e:
        logging.info(str(e) + "!")
        raise RuntimeError(f"Failed to create device nodes in the '{chroot}' chroot.") from e


def mount_dev_tmpfs(chroot: Chroot = Chroot.native()) -> None:
    """
    Mount tmpfs inside the chroot's dev folder to make sure we can create
    device nodes, even if the filesystem of the work folder does not support
    it.
    """
    # Do nothing when it is already mounted
    dev = chroot / "dev"
    if pmb.helpers.mount.ismount(dev):
        return

    # Create the $chroot/dev folder and mount tmpfs there
    pmb.helpers.run.root(["mkdir", "-p", dev])
    pmb.helpers.run.root(["mount", "-t", "tmpfs", "-o", "size=1M,noexec,dev", "tmpfs", dev])

    # Create /dev/{stdin,stdout,stderr}
    pmb.helpers.run.root(["ln", "-s", "/proc/self/fd/0", dev / "stdin"])
    pmb.helpers.run.root(["ln", "-s", "/proc/self/fd/1", dev / "stdout"])
    pmb.helpers.run.root(["ln", "-s", "/proc/self/fd/2", dev / "stderr"])

    # Create pts, shm folders and device nodes
    pmb.helpers.run.root(["mkdir", "-p", dev / "pts", dev / "shm"])
    pmb.helpers.run.root(
        ["mount", "-t", "tmpfs", "-o", "nodev,nosuid,noexec", "tmpfs", dev / "shm"]
    )

    # Mount devpts for PTY support
    pmb.helpers.run.root(["mount", "-t", "devpts", "-o", "ptmxmode=0666", "devpts", dev / "pts"])

    create_device_nodes(chroot)

    # Setup /dev/fd as a symlink
    pmb.helpers.run.root(["ln", "-sf", "/proc/self/fd", f"{dev}/"])


def mount(chroot: Chroot) -> None:
    if chroot.type == ChrootType.IMAGE and not pmb.mount.ismount(chroot.path):
        mount_chroot_image(chroot)

    # Mount tmpfs as the chroot's /dev
    mount_dev_tmpfs(chroot)

    # Get all mountpoints
    arch = chroot.arch
    channel = pmb.config.pmaports.read_config(pkgrepo_default_path())["channel"]
    mountpoints: dict[Path, Path] = {}
    for src_template, target_template in pmb.config.chroot_mount_bind.items():
        src_template = src_template.replace("$WORK", os.fspath(get_context().config.work))
        src_template = src_template.replace("$ARCH", str(arch))
        src_template = src_template.replace("$CHANNEL", channel)
        mountpoints[Path(src_template)] = Path(target_template)

    # Mount if necessary
    for source, target in mountpoints.items():
        target_outer = chroot / target
        if not pmb.helpers.mount.ismount(target_outer):
            pmb.helpers.mount.bind(source, target_outer)

    # Set up binfmt
    if not arch.cpu_emulation_required():
        return

    arch_qemu = arch.qemu_user()

    # mount --bind the qemu-user binary
    pmb.chroot.binfmt.register(arch)
    pmb.helpers.mount.bind_file(
        Chroot.native() / f"usr/bin/qemu-{arch_qemu}",
        chroot / f"usr/bin/qemu-{arch_qemu}-static",
        create_folders=True,
    )


def mount_native_into_foreign(chroot: Chroot) -> None:
    source = Chroot.native().path
    target = chroot / "native"
    pmb.helpers.mount.bind(source, target)

    musl = next(source.glob("lib/ld-musl-*.so.1")).name
    musl_link = chroot / "lib" / musl
    if not musl_link.is_symlink():
        pmb.helpers.run.root(["ln", "-s", "/native/lib/" + musl, musl_link])


def remove_mnt_pmbootstrap(chroot: Chroot) -> None:
    """
    Safely remove /mnt/pmbootstrap directories from the chroot, without
    running rm -r as root and potentially removing data inside the
    mountpoint in case it was still mounted (bug in pmbootstrap, or user
    ran pmbootstrap 2x in parallel). This is similar to running 'rm -r -d',
    but we don't assume that the host's rm has the -d flag (busybox does
    not).
    """
    mnt_dir = chroot / "mnt/pmbootstrap"

    if not mnt_dir.exists():
        return

    for path in [*mnt_dir.glob("*"), mnt_dir]:
        pmb.helpers.run.root(["rmdir", path])



# Copyright 2024 Caleb Connolly
# SPDX-License-Identifier: GPL-3.0-or-later

from .chroot import (
    Chroot as Chroot,
    ChrootType as ChrootType,
)
from .config import Config as Config
from .context import Context as Context



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.config.pmaports
from pmb.core.chroot import Chroot
from pmb.core.context import get_context


def variables(
    flavor: str | None,
    method: str,
    cmdline: str | None,
    no_reboot: bool | None,
    partition: str | None,
    resume: bool | None,
) -> dict[str, str | None]:
    device = get_context().config.device
    deviceinfo = pmb.parse.deviceinfo()
    cmdline_ = deviceinfo.kernel_cmdline or ""
    if cmdline:
        cmdline_ = cmdline

    flash_pagesize = deviceinfo.flash_pagesize

    # TODO Remove _partition_system deviceinfo support once pmaports has been
    # updated and minimum pmbootstrap version bumped.
    # See also https://gitlab.postmarketos.org/postmarketOS/pmbootstrap/-/issues/2243

    partition_kernel: str | None
    partition_vendor_boot: str | None
    partition_rootfs: str | None

    if method.startswith("fastboot"):
        partition_kernel = deviceinfo.flash_fastboot_partition_kernel or "boot"
        partition_vendor_boot = deviceinfo.flash_fastboot_partition_vendor_boot or "vendor_boot"
        partition_rootfs = (
            deviceinfo.flash_fastboot_partition_rootfs
            or deviceinfo.flash_fastboot_partition_system
            or "userdata"
        )
        partition_vbmeta = deviceinfo.flash_fastboot_partition_vbmeta or None
        partition_dtbo = deviceinfo.flash_fastboot_partition_dtbo or None
    # Require that the partitions are specified in deviceinfo for now
    elif method.startswith("rkdeveloptool"):
        partition_kernel = deviceinfo.flash_rk_partition_kernel or None
        partition_vendor_boot = deviceinfo.flash_rk_partition_vendor_boot or None
        partition_rootfs = (
            deviceinfo.flash_rk_partition_rootfs or deviceinfo.flash_rk_partition_system or None
        )
        partition_vbmeta = None
        partition_dtbo = None
    elif method.startswith("mtkclient"):
        partition_kernel = deviceinfo.flash_mtkclient_partition_kernel or "boot"
        partition_vendor_boot = deviceinfo.flash_mtkclient_partition_vendor_boot or "vendor_boot"
        partition_rootfs = deviceinfo.flash_mtkclient_partition_rootfs or "userdata"
        partition_vbmeta = deviceinfo.flash_mtkclient_partition_vbmeta or None
        partition_dtbo = deviceinfo.flash_mtkclient_partition_dtbo or None
    else:
        partition_kernel = deviceinfo.flash_heimdall_partition_kernel or "KERNEL"
        partition_vendor_boot = (
            deviceinfo.flash_heimdall_partition_vendor_boot or None
        )  # TODO: is there a default name?
        partition_rootfs = (
            deviceinfo.flash_heimdall_partition_rootfs
            or deviceinfo.flash_heimdall_partition_system
            or "SYSTEM"
        )
        partition_vbmeta = deviceinfo.flash_heimdall_partition_vbmeta or None
        partition_dtbo = deviceinfo.flash_heimdall_partition_dtbo or None

    if partition:
        # Only one operation is done at same time so it doesn't matter
        # sharing the arg
        partition_kernel = partition
        partition_vendor_boot = partition
        partition_rootfs = partition
        partition_vbmeta = partition
        partition_dtbo = partition

    dtb = deviceinfo.dtb + ".dtb"

    no_reboot_ = ""
    if no_reboot:
        no_reboot_ = "--no-reboot"

    resume_ = ""
    if resume:
        resume_ = "--resume"

    fvars = {
        "$BOOT": "/mnt/rootfs_" + device + "/boot",
        "$DTB": dtb,
        "$IMAGE_SPLIT_BOOT": "/home/pmos/rootfs/" + device + "-boot.img",
        "$IMAGE_SPLIT_ROOT": "/home/pmos/rootfs/" + device + "-root.img",
        "$IMAGE": "/home/pmos/rootfs/" + device + ".img",
        "$KERNEL_CMDLINE": cmdline_,
        "$PARTITION_KERNEL": partition_kernel,
        "$PARTITION_VENDOR_BOOT": partition_vendor_boot,
        "$PARTITION_INITFS": deviceinfo.flash_heimdall_partition_initfs or "RECOVERY",
        "$PARTITION_ROOTFS": partition_rootfs,
        "$PARTITION_VBMETA": partition_vbmeta,
        "$PARTITION_DTBO": partition_dtbo,
        "$FLASH_PAGESIZE": flash_pagesize,
        "$RECOVERY_ZIP": f"/mnt/{Chroot.buildroot(deviceinfo.arch)}"
        "/var/lib/postmarketos-android-recovery-installer"
        f"/pmos-{device}.zip",
        "$UUU_SCRIPT": f"/mnt/{Chroot.rootfs(deviceinfo.codename)}/usr/share/uuu/flash_script.lst",
        "$NO_REBOOT": no_reboot_,
        "$RESUME": resume_,
    }

    # Backwards compatibility with old mkinitfs (pma#660)
    pmaports_cfg = pmb.config.pmaports.read_config()
    if pmaports_cfg.get("supported_mkinitfs_without_flavors", False):
        fvars["$FLAVOR"] = ""
    else:
        fvars["$FLAVOR"] = f"-{flavor}" if flavor is not None else "-"

    return fvars



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import subprocess
from collections.abc import Sequence
from pathlib import Path
from typing import Literal, overload

import pmb.helpers.run_core
from pmb.types import (
    Env,
    PathString,
    RunOutputType,
    RunOutputTypeDefault,
    RunOutputTypePopen,
    RunReturnType,
)


def user(
    cmd: Sequence[PathString],
    working_dir: Path | None = None,
    output: RunOutputType = RunOutputTypeDefault.LOG,
    output_return: bool = False,
    check: bool | None = None,
    env: Env = {},
    sudo: bool = False,
) -> RunReturnType:
    """
    Run a command on the host system as user.

    :param env: dict of environment variables to be passed to the command, e.g.
                {"JOBS": "5"}

    See pmb.helpers.run_core.core() for a detailed description of all other
    arguments and the return value.
    """
    cmd_parts = [os.fspath(c) for c in cmd]

    # Readable log message (without all the escaping)
    msg = "% "
    for key, value in env.items():
        msg += f"{key}={value} "
    if working_dir is not None:
        msg += f"cd {os.fspath(working_dir)}; "
    msg += " ".join(cmd_parts)

    # Add environment variables and run
    env = env.copy()
    pmb.helpers.run_core.add_proxy_env_vars(env)
    if env:
        cmd_parts = ["sh", "-c", pmb.helpers.run_core.flat_cmd([cmd_parts], env=env)]
    return pmb.helpers.run_core.core(
        msg, cmd_parts, working_dir, output, output_return, check, sudo
    )


# FIXME: should probably use some kind of wrapper class / builder pattern for all these parameters...
def user_output(
    cmd: Sequence[PathString],
    working_dir: Path | None = None,
    output: RunOutputType = RunOutputTypeDefault.LOG,
    check: bool | None = None,
    env: Env = {},
    sudo: bool = False,
) -> str:
    ret = user(cmd, working_dir, output, output_return=True, check=check, env=env, sudo=sudo)
    if not isinstance(ret, str):
        raise TypeError("Expected str output, got " + str(ret))

    return ret


@overload
def root(
    cmd: Sequence[PathString],
    working_dir: Path | None = ...,
    output: RunOutputTypePopen = ...,
    output_return: Literal[False] = ...,
    check: bool | None = ...,
    env: Env = ...,
) -> subprocess.Popen: ...


@overload
def root(
    cmd: Sequence[PathString],
    working_dir: Path | None = ...,
    output: RunOutputTypeDefault = ...,
    output_return: Literal[False] = ...,
    check: bool | None = ...,
    env: Env = ...,
) -> int: ...


@overload
def root(
    cmd: Sequence[PathString],
    working_dir: Path | None = ...,
    output: RunOutputType = ...,
    output_return: Literal[True] = ...,
    check: bool | None = ...,
    env: Env = ...,
) -> str: ...


def root(
    cmd: Sequence[PathString],
    working_dir: Path | None = None,
    output: RunOutputType = RunOutputTypeDefault.LOG,
    output_return: bool = False,
    check: bool | None = None,
    env: Env = {},
) -> RunReturnType:
    """
    Run a command on the host system as root with sudo, doas, or run0.

    :param env: dict of environment variables to be passed to the command, e.g.
                {"JOBS": "5"}

    See pmb.helpers.run_core.core() for a detailed description of all other
    arguments and the return value.
    """
    env = env.copy()
    pmb.helpers.run_core.add_proxy_env_vars(env)

    if env:
        cmd = ["sh", "-c", pmb.helpers.run_core.flat_cmd([cmd], env=env)]
    cmd = pmb.config.sudo(cmd)

    return user(cmd, working_dir, output, output_return, check, env, True)



# Copyright 2026 Oliver Smith, Paul Adam
# SPDX-License-Identifier: GPL-3.0-or-later
import json
from pathlib import Path

import pmb.parse.apkindex


def apkindex_parse(apkindex_path: Path, package: str | list[str]) -> None:
    result = pmb.parse.apkindex.parse(apkindex_path)
    if package:
        if package not in result:
            raise RuntimeError(f"Package not found in the APKINDEX: {package}")
        print(json.dumps(result[package], indent=4, default=vars))
    else:
        print(json.dumps(result, indent=4, default=vars))



# Copyright 2024 Stefan Hansson
# SPDX-License-Identifier: GPL-3.0-or-later

import pmb.helpers.pkgrel_bump


def pkgver_bump(packages: list[str]) -> None:
    # Each package must exist
    for package in packages:
        pmb.helpers.pmaports.find(package)

    for package in packages:
        pmb.helpers.pkgrel_bump.package(package, bump_type=pmb.helpers.pkgrel_bump.BumpType.PKGVER)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os

import pmb.chroot.apk
import pmb.install
from pmb.core import Chroot
from pmb.core.context import get_context
from pmb.helpers import logging


def kernel_flavor_installed(chroot: Chroot, autoinstall: bool = True) -> str | None:
    """
    Get installed kernel flavor. Optionally install the device's kernel
    beforehand.

    :param suffix: the chroot suffix, e.g. "native" or "rootfs_qemu-amd64"
    :param autoinstall: install the device's kernel if it is not installed
    :returns: * string with the installed kernel flavor,
                e.g. ["postmarketos-qcom-sdm845"]
              * None if no kernel is installed
    """
    # Automatically install the selected kernel
    if autoinstall:
        if not chroot.is_mounted():
            pmb.chroot.init(chroot)
        config = get_context().config
        packages = [f"device-{config.device}", *pmb.install.get_kernel_package(config)]
        pmb.chroot.apk.install(packages, chroot)

    glob_result = list((chroot / "usr/share/kernel").glob("*"))

    # There should be only one directory here
    return glob_result[0].name if glob_result else None


def copy_xauthority(chroot: Chroot) -> None:
    """
    Copy the host system's Xauthority file to the pmos user inside the chroot,
    so we can start X11 applications from there.
    """
    # Check $DISPLAY
    logging.info("(native) copy host Xauthority")
    if not os.environ.get("DISPLAY"):
        raise RuntimeError(
            "Your $DISPLAY variable is not set. If you have an"
            " X11 server running as your current user, try"
            " 'export DISPLAY=:0' and run your last"
            " pmbootstrap command again."
        )

    # Check $XAUTHORITY
    original = os.environ.get("XAUTHORITY")
    if not original:
        original = os.path.join(os.environ["HOME"], ".Xauthority")
    if not os.path.exists(original):
        raise RuntimeError(
            "Could not find your Xauthority file, try to export"
            " your $XAUTHORITY correctly. Looked here: " + original
        )

    # Copy to chroot and chown
    copy = chroot / "home/pmos/.Xauthority"
    if os.path.exists(copy):
        pmb.helpers.run.root(["rm", copy])
    pmb.helpers.run.root(["cp", original, copy])
    pmb.chroot.root(["chown", "pmos:pmos", "/home/pmos/.Xauthority"])



# Copyright 2024 Stefan Hansson
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

import pmb.aportgen
from pmb.helpers import logging


def aportgen(package_list: list[str], fork_alpine: bool, fork_alpine_retain_branch: bool) -> None:
    for package in package_list:
        logging.info(f"Generate aport: {package}")
        pmb.aportgen.generate(package, fork_alpine, fork_alpine_retain_branch)



# Copyright 2024 Stefan "Newbyte" Hansson
# SPDX-License-Identifier: GPL-3.0-or-later


class BuildFailedError(Exception):
    """
    Exception to be raised when pmbootstrap fails to build a package. This is handled
    separately from NonBugError as it needs to be treated differently as we want to hint
    to users that they can check the log for more information when a build fails.
    """


class CommandFailedError(Exception):
    """
    Exception to be raised when a command pmbootstrap has launched via shell fails. Should only be
    handled on a selective basis.
    """


class NonBugError(Exception):
    """
    Exception which originates from a problem not caused by pmbootstrap's code. This
    could for example be raised if there is an error in a package pmbootstrap is
    interacting with in some way.
    """


class PackagingError(Exception):
    """
    Exception which at least very likely originates from a problem in the pmaports/aports packaging.
    This could for instance be due to a package specifying a path to a file that no longer exists or
    mkinitfs failing to run.
    """



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

from pathlib import PosixPath

import pmb.chroot.apk
import pmb.chroot.initfs
import pmb.chroot.other
import pmb.config
import pmb.flasher
import pmb.helpers.frontend
import pmb.helpers.mount
import pmb.parse.kconfig
from pmb.core import Chroot, ChrootType
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError
from pmb.parse.deviceinfo import Deviceinfo
from pmb.types import RunOutputTypeDefault


def kernel(
    deviceinfo: Deviceinfo,
    method: str,
    boot: bool = False,
    autoinstall: bool = False,
    cmdline: str | None = None,
    no_reboot: bool | None = None,
    partition: str | None = None,
    resume: bool | None = None,
) -> None:
    # Rebuild the initramfs, just to make sure (see #69)
    flavor = pmb.helpers.frontend._parse_flavor(deviceinfo.codename, autoinstall)
    if autoinstall:
        pmb.chroot.initfs.build(Chroot(ChrootType.ROOTFS, deviceinfo.codename))

    # deviceinfo.header_version >= 3 == GKI == has / needs vendor_boot
    is_gki = deviceinfo.header_version and deviceinfo.header_version >= 3

    # Check kernel config
    pmb.parse.kconfig.check(flavor, must_exist=False)

    # Do not allow header_version >= 3 for non-fastboot, since the correct way to flash
    # vendor_boot with other flash methods is unknown as of 2025-08.
    if not deviceinfo.flash_method == "fastboot" and (
        deviceinfo.header_version and deviceinfo.header_version >= 3
    ):
        raise NonBugError(
            "deviceinfo_header_version' <= 2 is currently only supported with fastboot. See also:"
            " <https://postmarketos.org/deviceinfo>"
        )

    # Generate the paths and run the flasher
    if boot:
        logging.info("(native) boot " + flavor + " kernel")
        pmb.flasher.run(
            deviceinfo,
            method,
            "boot_gki" if is_gki else "boot",
            flavor,
            cmdline=cmdline,
            no_reboot=no_reboot,
            partition=partition,
            resume=resume,
        )
    else:
        logging.info("(native) flash kernel " + flavor)
        pmb.flasher.run(
            deviceinfo,
            method,
            "flash_kernel",
            flavor,
            cmdline=cmdline,
            no_reboot=no_reboot,
            partition=partition,
            resume=resume,
        )
        if is_gki:
            pmb.flasher.run(
                deviceinfo,
                method,
                "flash_vendorboot",
                flavor,
                cmdline=cmdline,
                no_reboot=no_reboot,
                partition=partition,
                resume=resume,
            )

    logging.info("You will get an IP automatically assigned to your USB interface shortly.")
    logging.info("Then you can connect to your device using ssh after pmOS has booted:")
    logging.info(f"ssh {get_context().config.user}@{pmb.config.default_ip}")
    logging.info(
        "NOTE: If you enabled full disk encryption, you should make"
        " sure that Unl0kr has been properly configured for your"
        " device"
    )


def list_flavors(device: str) -> None:
    chroot = Chroot(ChrootType.ROOTFS, device)
    logging.info(f"({chroot}) installed kernel flavors:")
    logging.info("* " + str(pmb.chroot.other.kernel_flavor_installed(chroot)))


def rootfs(
    deviceinfo: Deviceinfo,
    method: str,
    cmdline: str | None = None,
    no_reboot: bool | None = None,
    partition: str | None = None,
    resume: bool | None = None,
) -> None:
    # Generate rootfs, install flasher
    suffix = ".img"
    if pmb.config.flashers.get(method, {}).get("split", False):
        suffix = "-root.img"

    img_path = Chroot.native() / "home/pmos/rootfs" / f"{deviceinfo.codename}{suffix}"
    if not img_path.exists():
        raise NonBugError(
            "The rootfs has not been generated yet, please run 'pmbootstrap install' first."
        )

    # Do not flash if using fastboot & image is too large
    if method.startswith("fastboot") and deviceinfo.flash_fastboot_max_size:
        img_size = img_path.stat().st_size / 1024**2
        max_size = int(deviceinfo.flash_fastboot_max_size)
        if img_size > max_size:
            raise RuntimeError("The rootfs is too large for fastboot to flash.")

    # Run the flasher
    logging.info("(native) flash rootfs image")
    pmb.flasher.run(
        deviceinfo,
        method,
        "flash_rootfs",
        cmdline=cmdline,
        no_reboot=no_reboot,
        partition=partition,
        resume=resume,
    )


def list_devices(
    deviceinfo: Deviceinfo,
    method: str,
    cmdline: str | None = None,
    no_reboot: bool | None = None,
    partition: str | None = None,
    resume: bool | None = None,
) -> None:
    # pmb.flasher.run provides user-facing error messages for unsupported
    # flags, hence why we pass them here despite them being unused
    pmb.flasher.run(
        deviceinfo,
        method,
        "list_devices",
        cmdline=cmdline,
        no_reboot=no_reboot,
        partition=partition,
        resume=resume,
    )


def sideload(
    deviceinfo: Deviceinfo,
    method: str,
    cmdline: str | None = None,
    no_reboot: bool | None = None,
    partition: str | None = None,
    resume: bool | None = None,
) -> None:
    # Install depends
    pmb.flasher.install_depends(method)

    # Mount the buildroot
    chroot = Chroot.buildroot(deviceinfo.arch)
    mountpoint = "mnt" / PosixPath(chroot.__str__())
    pmb.helpers.mount.bind(chroot.path, Chroot.native().path / mountpoint)

    # Missing recovery zip error
    if not (
        Chroot.native().path
        / mountpoint
        / "var/lib/postmarketos-android-recovery-installer"
        / f"pmos-{deviceinfo.codename}.zip"
    ).exists():
        raise RuntimeError(
            "The recovery zip has not been generated yet,"
            " please run 'pmbootstrap install' with the"
            " '--android-recovery-zip' parameter first!"
        )

    pmb.flasher.run(
        deviceinfo,
        method,
        "sideload",
        cmdline=cmdline,
        no_reboot=no_reboot,
        partition=partition,
        resume=resume,
    )


def flash_lk2nd(
    deviceinfo: Deviceinfo,
    method: str,
    cmdline: str | None = None,
    no_reboot: bool | None = None,
    partition: str | None = None,
    resume: bool | None = None,
) -> None:
    if method == "fastboot":
        # In the future this could be expanded to use "fastboot flash lk2nd $img"
        # which reflashes/updates lk2nd from itself. For now let the user handle this
        # manually since supporting the codepath with heimdall requires more effort.
        pmb.flasher.init(deviceinfo.codename, method)
        logging.info("(native) checking current fastboot product")
        output = pmb.chroot.root(
            ["fastboot", "getvar", "product"],
            output=RunOutputTypeDefault.INTERACTIVE,
            output_return=True,
        )
        # Variable "product" is e.g. "LK2ND_MSM8974" or "lk2nd-msm8226" depending
        # on the lk2nd version.
        if "lk2nd" in output.lower():
            raise RuntimeError(
                "You are currently running lk2nd. Please reboot into the regular"
                " bootloader mode to re-flash lk2nd."
            )

    # Get the lk2nd package (which is a dependency of the device package)
    device_pkg = f"device-{deviceinfo.codename}"
    apkbuild = pmb.helpers.pmaports.get(device_pkg)
    if not apkbuild:
        raise RuntimeError(f"Failed to find {device_pkg} in pmaports")
    lk2nd_pkg = None
    for dep in apkbuild["depends"]:
        if dep.startswith("lk2nd"):
            lk2nd_pkg = dep
            break

    # If not found, also check subpackages
    if not lk2nd_pkg:
        for subpackage in apkbuild["subpackages"]:
            for dep in apkbuild["subpackages"][subpackage]["depends"]:
                if dep.startswith("lk2nd"):
                    lk2nd_pkg = dep
                    break

    if not lk2nd_pkg:
        raise RuntimeError(f"{device_pkg} does not depend on any lk2nd package")

    suffix = Chroot(ChrootType.ROOTFS, deviceinfo.codename)
    pmb.chroot.apk.install([lk2nd_pkg], suffix)

    logging.info("(native) flash lk2nd image")
    pmb.flasher.run(
        deviceinfo,
        method,
        "flash_lk2nd",
        cmdline=cmdline,
        no_reboot=no_reboot,
        partition=partition,
        resume=resume,
    )



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.config
import pmb.helpers.pmaports
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.meta import Cache
from pmb.types import Apkbuild, CrossCompile


@Cache("package")
def arch(package: str | Apkbuild) -> Arch:
    """
    Find a good default in case the user did not specify for which architecture
    a package should be built.

    :param package: The name of the package or parsed APKBUILD

    :returns: Arch object. Preferred order, depending
              on what is supported by the APKBUILD:
              * native arch
              * device arch (this will be preferred instead if build_default_device_arch is true)
              * first arch in the APKBUILD
    """
    pkgname = package["pkgname"] if isinstance(package, dict) else package
    aport = pmb.helpers.pmaports.find(pkgname)

    apkbuild = pmb.parse.apkbuild(aport) if isinstance(package, str) else package
    arches = apkbuild["arch"]
    deviceinfo = pmb.parse.deviceinfo()

    if get_context().config.build_default_device_arch:
        preferred_arch = deviceinfo.arch
        preferred_arch_2nd = Arch.native()
    else:
        preferred_arch = Arch.native()
        preferred_arch_2nd = deviceinfo.arch

    # TODO: Ideally we wouldn't convert preferred_arch and preferred_arch_2nd to strings here and
    # instead be able to parse apkbuild["arch"] in a properly typed way.
    if "noarch" in arches or "all" in arches or str(preferred_arch) in arches:
        return preferred_arch

    if str(preferred_arch_2nd) in arches:
        return preferred_arch_2nd

    try:
        arch_str = apkbuild["arch"][0]
        return Arch.from_str(arch_str) if arch_str else Arch.native()
    except IndexError:
        return Arch.native()


def crosscompile(apkbuild: Apkbuild, arch: Arch) -> CrossCompile:
    """Decide the type of compilation necessary to build a given APKBUILD."""
    if not get_context().cross:
        return CrossCompile.QEMU_ONLY
    if not arch.cpu_emulation_required():
        return CrossCompile.UNNECESSARY
    if "pmb:cross-native" in apkbuild["options"]:
        return CrossCompile.CROSS_NATIVE
    # In case we end up being requested to cross-compile a "noarch" package,
    # default to cross-native2 if the package has "noarch" since that implies
    # it won't run a compiler. This can happen if e.g. the --arch argument is
    # set to a foreign architecture when calling the build subcommand or as
    # part of the install subcommand when building an image for a foreign
    # architecture.
    if (
        arch.is_native()
        or "pmb:cross-native2" in apkbuild["options"]
        or apkbuild["arch"] == ["noarch"]
    ):
        return CrossCompile.CROSS_NATIVE2
    if "!pmb:crossdirect" in apkbuild["options"]:
        return CrossCompile.QEMU_ONLY
    return CrossCompile.CROSSDIRECT



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
from pmb.chroot.init import (
    init as init,
    init_keys as init_keys,
)
from pmb.chroot.mount import (
    mount as mount,
    mount_native_into_foreign as mount_native_into_foreign,
    remove_mnt_pmbootstrap as remove_mnt_pmbootstrap,
)
from pmb.chroot.run import (
    root as root,
    rootm as rootm,
    user as user,
    user_exists as user_exists,
    userm as userm,
)
from pmb.chroot.shutdown import shutdown as shutdown
from pmb.chroot.zap import (
    del_chroot as del_chroot,
    zap as zap,
)



# Copyright 2023 Johannes Marbach, Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import shlex
from collections.abc import Sequence
from pathlib import Path
from typing import Literal

import pmb.config.pmaports
import pmb.helpers.cli
import pmb.helpers.repo
import pmb.helpers.run
import pmb.helpers.run_core
import pmb.parse.version
from pmb.core.arch import Arch
from pmb.core.chroot import Chroot
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.meta import Cache
from pmb.types import PathString, RunOutputTypePopen


@Cache("root", "user_repository", mirrors_exclude=[])
def update_repository_list(
    root: Path,
    user_repository: bool | Path = False,
    mirrors_exclude: list[str] | Literal[True] = [],
    check: bool = False,
) -> None:
    """
    Update /etc/apk/repositories, if it is outdated (when the user changed the
    --mirror-alpine or --mirror-pmOS parameters).

    :param root: the root directory to operate on
    :param mirrors_exclude: mirrors to exclude from the repository list
    :param check: This function calls it self after updating the
                  /etc/apk/repositories file, to check if it was successful.
                  Only for this purpose, the "check" parameter should be set to
                  True.
    """
    # Read old entries or create folder structure
    path = root / "etc/apk/repositories"
    lines_old: list[str] = []
    if path.exists():
        # Read all old lines
        lines_old = []
        with path.open() as handle:
            for line in handle:
                lines_old.append(line[:-1])
    else:
        pmb.helpers.run.root(["mkdir", "-p", path.parent])

    user_repo_dir: Path | None
    if isinstance(user_repository, Path):
        user_repo_dir = user_repository
    else:
        user_repo_dir = Path("/mnt/pmbootstrap/packages") if user_repository else None

    # Up to date: Save cache, return
    lines_new = pmb.helpers.repo.urls(
        user_repository=user_repo_dir, mirrors_exclude=mirrors_exclude
    )
    if lines_old == lines_new:
        return

    # Check phase: raise error when still outdated
    if check:
        raise RuntimeError(f"Failed to update: {path}")

    # Update the file
    logging.debug(f"({root.name}) update /etc/apk/repositories")
    if path.exists():
        pmb.helpers.run.root(["rm", path])
    for line in lines_new:
        pmb.helpers.run.root(["sh", "-c", f"echo {shlex.quote(line)} >> {path}"])
    update_repository_list(
        root,
        user_repository=user_repository,
        mirrors_exclude=mirrors_exclude,
        check=True,
    )


def _prepare_fifo() -> Path:
    """
    Prepare the progress fifo for reading / writing.

    :param chroot: whether to run the command inside the chroot or on the host
    :param suffix: chroot suffix. Only applies if the "chroot" parameter is
                   set to True.
    :returns: A tuple consisting of the path to the fifo as needed by apk to
              write into it (relative to the chroot, if applicable) and the
              path of the fifo as needed by cat to read from it (always
              relative to the host)
    """
    pmb.helpers.run.root(["mkdir", "-p", get_context().config.work / "tmp"])
    fifo = get_context().config.work / "tmp/apk_progress_fifo"
    if os.path.exists(fifo):
        pmb.helpers.run.root(["rm", "-f", fifo])

    pmb.helpers.run.root(["mkfifo", fifo])
    return fifo


def _create_command_with_progress(command: Sequence[str], fifo: Path) -> list[str]:
    """
    Build a full apk command from a subcommand, set up to redirect progress into a fifo.

    :param command: apk subcommand in list form
    :param fifo: path of the fifo
    :returns: full command in list form
    """
    flags = ["--progress-fd", "3"]
    command_full = [command[0], *flags, *command[1:]]
    command_flat = pmb.helpers.run_core.flat_cmd([command_full])
    command_flat = f"exec 3>{fifo}; {command_flat}"
    return ["sh", "-c", command_flat]


def _compute_progress(line: str) -> float:
    """
    Compute the progress as a number between 0 and 1.

    :param line: line as read from the progress fifo
    :returns: progress as a number between 0 and 1
    """
    if not line:
        return 1
    cur_tot = line.split(" ")[0].rstrip().split("/")
    if len(cur_tot) != 2:
        return 0
    cur = float(cur_tot[0])
    tot = float(cur_tot[1])
    return cur / tot if tot > 0 else 0


def _apk_with_progress(command: Sequence[str]) -> None:
    """
    Run an apk subcommand while printing a progress bar to STDOUT.

    :param command: apk subcommand in list form
    :raises RuntimeError: when the apk command fails
    """
    fifo = _prepare_fifo()
    command_with_progress = _create_command_with_progress(command, fifo)
    log_msg = " ".join(command)
    with (
        pmb.helpers.run.root(["cat", fifo], output=RunOutputTypePopen.PIPE) as p_cat,
        pmb.helpers.run.root(command_with_progress, output=RunOutputTypePopen.BACKGROUND) as p_apk,
    ):
        while p_apk.poll() is None:
            p_cat_stdout = p_cat.stdout
            if p_cat_stdout is None:
                raise RuntimeError("cat process had no stdout?")
            line = p_cat_stdout.readline().decode("utf-8")
            progress = _compute_progress(line)
            pmb.helpers.cli.progress_print(progress)
        pmb.helpers.cli.progress_flush()
        pmb.helpers.run_core.check_return_code(p_apk.returncode, log_msg)


def _prepare_cmd(command: Sequence[PathString], chroot: Chroot | None) -> list[str]:
    """
    Prepare the apk command.

    Returns a tuple of the first part of the command with generic apk flags, and the second part
    with the subcommand and its arguments.
    """
    config = get_context().config
    # Our _apk_with_progress() wrapper also need --no-progress, since all that does is
    # prevent apk itself from rendering progress bars. We instead want it to tell us
    # the progress so we can render it. So we always set --no-progress.
    command_: list[str] = [str(config.work / "apk.static"), "--no-progress"]
    if chroot:
        cache_dir = config.work / f"cache_apk_{chroot.arch}"
        command_.extend(
            [
                "--root",
                str(chroot.path),
                "--arch",
                str(chroot.arch),
            ]
        )

        if os.getenv("PMB_APK_NO_CACHE") == "1":
            command_.extend(["--no-cache"])
        else:
            command_.extend(["--cache-dir", str(cache_dir)])

        local_repos = pmb.helpers.repo.urls(
            user_repository=config.work / "packages", mirrors_exclude=True
        )
        for repo in local_repos:
            command_.extend(["--repository", repo])
    if get_context().offline:
        command_.append("--no-network")

    for c in command:
        command_.append(os.fspath(c))

        # Always be non-interactive
        if c == "add":
            command_.append("--no-interactive")

    return command_


def run(command: Sequence[PathString], chroot: Chroot, with_progress: bool = True) -> None:
    """
    Run an apk subcommand.

    :param command: apk subcommand in list form
    :param with_progress: whether to print a progress bar
    :param chroot: chroot to run the command in
    :raises RuntimeError: when the apk command fails
    """
    command_ = _prepare_cmd(command, chroot)

    # Sanity checks. We should avoid accidentally writing to
    # /var/cache/apk on the host!
    if "add" in command:
        if "--no-interactive" not in command_:
            raise RuntimeError(
                "Encountered an 'apk add' command without --no-interactive! This is a bug."
            )
        if os.getenv("PMB_APK_NO_CACHE") == "1" and "--no-cache" not in command_:
            raise RuntimeError(
                "Encountered an 'apk add' command without --no-cache! This is a bug."
            )
        if os.getenv("PMB_APK_NO_CACHE") != "1" and "--cache-dir" not in command_:
            raise RuntimeError(
                "Encountered an 'apk add' command without --cache-dir! This is a bug."
            )

    if with_progress:
        _apk_with_progress(command_)
    else:
        pmb.helpers.run.root(command_)


def cache_clean(arch: Arch) -> None:
    """Clean the APK cache for a specific architecture."""
    work = get_context().config.work
    cache_dir = work / f"cache_apk_{arch}"
    if not cache_dir.exists():
        return

    # The problem here is that apk's "cache clean" command really
    # expects to be run against an apk-managed rootfs and will return
    # errors if it can't access certain paths (even though it will
    # actually clean the cache like we want).
    # We could just ignore the return value, but then we wouldn't know
    # if something actually went wrong with apk...
    # So we do this dance of creating a rootfs with only the files that
    # APK needs to be happy
    tmproot = work / "tmp_apk_root"  # pmb#2491: not using tmp/apk_root
    if not (tmproot / "etc/apk/repositories").exists():
        tmproot.mkdir(exist_ok=True)
        (tmproot / "var/cache").mkdir(exist_ok=True, parents=True)
        (tmproot / "etc/apk").mkdir(exist_ok=True, parents=True)
        (tmproot / "lib/apk/db").mkdir(exist_ok=True, parents=True)

        (tmproot / "etc/apk/world").touch(exist_ok=True)
        (tmproot / "lib/apk/db/installed").touch(exist_ok=True)
        (tmproot / "lib/apk/db/triggers").touch(exist_ok=True)

        (tmproot / "etc/apk/keys").symlink_to(work / "config_apk_keys")

        # Our fake rootfs needs a valid repositories file for apk
        # to have something to compare the cache against
        update_repository_list(tmproot, user_repository=work / "packages")

    # Point our tmproot cache dir to the real cache dir
    # this is much simpler than passing --cache-dir to apk
    # since even with that flag apk will also check it's
    # "static cache".
    (tmproot / "var/cache/apk").unlink(missing_ok=True)
    (tmproot / "var/cache/apk").symlink_to(cache_dir)

    command: list[PathString] = [
        "-v",
        "--root",
        tmproot,
        "--arch",
        str(arch),
    ]

    command += ["cache", "clean"]
    command_ = _prepare_cmd(command, None)

    pmb.helpers.apk_static.init()
    pmb.helpers.run.root(command_)


def check_outdated(version_installed: str) -> None:
    """
    Check if the provided alpine version is outdated.

    This depends on the alpine mirrordir (edge, v3.12, ...) related to currently checked out
    pmaports branch.

    :param version_installed: currently installed apk version, e.g. "2.12.1-r0"
    :raises: RuntimeError if the version is outdated
    """
    channel_cfg = pmb.config.pmaports.read_config_channel()
    mirrordir_alpine = channel_cfg["mirrordir_alpine"]
    version_min = pmb.config.apk_tools_min_version[mirrordir_alpine]

    if pmb.parse.version.compare(version_installed, version_min) >= 0:
        return

    raise RuntimeError(
        f"Found an outdated version of the 'apk' package manager ({version_installed}, expected at least: {version_min})"
    )



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import time
from pathlib import Path

import pmb.helpers.pmaports
import pmb.helpers.run


def replace(path: Path, old: str, new: str) -> None:
    text = ""
    with path.open("r", encoding="utf-8") as handle:
        text = handle.read()

    text = text.replace(old, new)

    with path.open("w", encoding="utf-8") as handle:
        handle.write(text)


def replace_apkbuild(pkgname: str, key: str, new: int | str, in_quotes: bool = False) -> None:
    """
    Replace one key=value line in an APKBUILD and verify it afterwards.

    :param pkgname: package name, e.g. "hello-world"
    :param key: key that should be replaced, e.g. "pkgver"
    :param new: new value
    :param in_quotes: expect the value to be in quotation marks ("")
    """
    # Read old value
    path = pmb.helpers.pmaports.find(pkgname) / "APKBUILD"
    apkbuild = pmb.parse.apkbuild(path)
    old = apkbuild[key]

    # Prepare old/new strings
    if in_quotes:
        line_old = f'{key}="{old}"'
        line_new = f'{key}="{new}"'
    else:
        line_old = f"{key}={old}"
        line_new = f"{key}={new}"

    # Replace
    replace(path, "\n" + line_old + "\n", "\n" + line_new + "\n")

    # Verify
    pmb.parse.apkbuild.cache_clear()
    apkbuild = pmb.parse.apkbuild(path)
    if apkbuild[key] != str(new):
        raise RuntimeError(
            f"Failed to set '{key}' for pmaport '{pkgname}'. Make sure"
            f" that there's a line with exactly the string '{line_old}'"
            f" and nothing else in: {path}"
        )


def is_up_to_date(
    path_sources: list[Path], path_target: Path | None = None, lastmod_target: float | None = None
) -> bool:
    """
    Check if a file is up-to-date by comparing the last modified timestamps.

    (just like make does it).

    :param path_sources: list of full paths to the source files
    :param path_target: full path to the target file
    :param lastmod_target: the timestamp of the target file. specify this as
                           alternative to specifying path_target.
    """
    if path_target and lastmod_target:
        raise RuntimeError("Specify path_target *or* lastmod_target, not both!")

    lastmod_source = None
    for path_source in path_sources:
        lastmod = os.path.getmtime(path_source)
        if not lastmod_source or lastmod > lastmod_source:
            lastmod_source = lastmod

    if path_target:
        lastmod_target = os.path.getmtime(path_target)

    if lastmod_target is None or lastmod_source is None:
        raise AssertionError

    return lastmod_target >= lastmod_source


def is_older_than(path: Path, seconds: int) -> bool:
    """Check if a single file is older than a given amount of seconds."""
    if not os.path.exists(path):
        return True
    lastmod = os.path.getmtime(path)
    return lastmod + seconds < time.time()


def symlink(file: Path, link: Path) -> None:
    """Check if the symlink is already present, otherwise create it."""
    if os.path.exists(link):
        if os.path.islink(link) and os.path.realpath(os.readlink(link)) == os.path.realpath(file):
            return
        raise RuntimeError(f"File exists: {link}")
    elif link.is_symlink():
        link.unlink()

    # Create the symlink
    pmb.helpers.run.user(["ln", "-s", file, link])



# Copyright 2023 Dylan Van Assche
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.helpers.pmaports
from pmb.core import Config
from pmb.helpers import logging


def get_groups(config: Config) -> list[str]:
    """
    Get all groups to which the user additionally must be added.
    The list of groups are listed in _pmb_groups of the UI and
    UI-extras package.

    :returns: list of groups, e.g. ["feedbackd", "udev"]
    """
    ret: list[str] = []
    if config.ui == "none":
        return ret

    # UI package
    meta = f"postmarketos-ui-{config.ui}"
    apkbuild = pmb.helpers.pmaports.get(meta)
    groups = apkbuild["_pmb_groups"]
    if groups:
        logging.debug(f"{meta}: install _pmb_groups: {', '.join(groups)}")
        ret += groups

    # UI-extras subpackage
    meta_extras = f"{meta}-extras"
    if config.ui_extras and meta_extras in apkbuild["subpackages"]:
        groups = apkbuild["subpackages"][meta_extras]["_pmb_groups"]
        if groups:
            logging.debug(f"{meta_extras}: install _pmb_groups: {', '.join(groups)}")
            ret += groups

    return ret



# Copyright 2025 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import glob
import logging
import os
from pathlib import Path

import pmb.build
import pmb.helpers.package
import pmb.helpers.pmaports
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.helpers.devices import DeviceCategory, get_device_category_by_apkbuild_path
from pmb.meta import Cache
from pmb.types import WithExtraRepos


@Cache("repo")
def is_abuild_forked(repo: str | None) -> bool:
    """Check if abuild is forked to make sure we build it first (pmb#2401)"""
    with_extra_repos: WithExtraRepos

    if repo == "systemd":
        with_extra_repos = "enabled"
    elif repo is None:
        with_extra_repos = "disabled"
    else:
        raise RuntimeError(f"Unexpected repo value: {repo}")

    return bool(pmb.helpers.pmaports.find("abuild", False, False, with_extra_repos))


def generate(arch: Arch) -> list[dict[str, list[str] | str | None]]:
    """
    Get packages that need to be built, with all their dependencies. Include
       packages from extra-repos, no matter if systemd is enabled or not. This
       is used by bpo to fill its package database.

    :param arch: architecture (e.g. "armhf")
    :returns: a list like the following:
        [{"pkgname": "hello-world", "repo": None, "version": "1-r4"},
        {"pkgname": "package-depending-on-hello-world", "version": "0.5-r0", "repo": None}]
    """
    ret = []

    for pmaports_dir in get_context().config.aports:
        pattern = os.path.join(pmaports_dir, "**/*/APKBUILD")

        for apkbuild_path_str in glob.glob(pattern, recursive=True):
            apkbuild_path = Path(apkbuild_path_str)
            pkgname = apkbuild_path.parent.name
            package = pmb.parse.apkbuild(apkbuild_path)
            version = f"{package['pkgver']}-r{package['pkgrel']}"

            if not pmb.helpers.pmaports.check_arches(package["arch"], arch):
                continue

            relpath = apkbuild_path.relative_to(pmaports_dir)
            if (
                relpath.parts[0] == "device"
                and get_device_category_by_apkbuild_path(relpath) == DeviceCategory.ARCHIVED
            ):
                logging.debug(f"skipping archived device: {relpath}")
                continue

            repo = relpath.parts[1] if relpath.parts[0] == "extra-repos" else None

            depends = []
            dep_fields = ["depends", "makedepends", "checkdepends"]
            for dep_field in dep_fields:
                for dep in package[dep_field]:
                    if dep.startswith("!"):
                        continue

                    dep_data = pmb.helpers.package.get(
                        dep, arch, must_exist=False, try_other_arches=False
                    )
                    if not dep_data:
                        logging.warning(f"WARNING: {pkgname}: failed to resolve dependency '{dep}'")
                        # Can't replace potential subpkgname
                        if dep != pkgname and dep not in depends:
                            depends += [dep]
                        continue
                    dep_pkgname = dep_data.pkgname
                    if dep_pkgname != pkgname and dep_pkgname not in depends:
                        depends += [dep_pkgname]

            # Add abuild to depends if needed
            if pkgname != "abuild" and is_abuild_forked(repo):
                depends = ["abuild", *depends]

            depends = sorted(depends)

            ret += [
                {
                    "pkgname": pkgname,
                    "repo": repo,
                    "version": version,
                    "depends": depends,
                }
            ]

    # "or -1" is needed for mypy
    # https://github.com/python/mypy/issues/9765#issuecomment-1238263745
    return sorted(ret, key=lambda d: d.get("pkgname") or -1)



# Copyright 2023 Pablo Castellano, Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import re
import shutil
import signal
import subprocess
from collections.abc import Sequence
from pathlib import Path
from types import FrameType

import pmb.chroot
import pmb.chroot.apk
import pmb.helpers.run
import pmb.install.losetup
import pmb.parse.cpuinfo
from pmb.core import Chroot, ChrootType
from pmb.core.arch import Arch
from pmb.core.config import Config
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError
from pmb.types import Env, PathString, RunOutputTypeDefault


def system_image(device: str) -> Path:
    """
    Returns path to rootfs for specified device. In case that it doesn't
    exist, raise and exception explaining how to generate it.
    """
    path = Chroot.native() / "home/pmos/rootfs" / f"{device}.img"
    if not path.exists():
        logging.debug(f"Could not find rootfs: {path}")
        raise NonBugError(
            "The rootfs has not been generated yet, please run 'pmbootstrap install' first."
        )
    return path


def create_second_storage(size: str, device: str) -> Path:
    """
    Generate a second storage image if it does not exist.

    :returns: path to the image or None

    """
    path = Chroot.native() / "home/pmos/rootfs" / f"{device}-2nd.img"
    pmb.helpers.run.root(["touch", path])
    pmb.helpers.run.root(["chmod", "a+w", path])
    resize_image(size, path)
    return path


def which_qemu(arch: Arch) -> str:
    """Finds the qemu executable or raises an exception otherwise"""
    executable = "qemu-system-" + arch.qemu_system()
    if shutil.which(executable):
        return executable
    else:
        raise NonBugError(
            f"Could not find the '{executable}' executable in your PATH."
            "  Please install it in order to run qemu."
        )


def create_gdk_loader_cache() -> Path:
    """
    Create a gdk loader cache that can be used for running GTK UIs outside of
    the chroot.
    """
    gdk_cache_dir = Path("/usr/lib/gdk-pixbuf-2.0/2.10.0/")
    custom_cache_path = gdk_cache_dir / "loaders-pmos-chroot.cache"
    chroot_native = Chroot.native()
    if (chroot_native / custom_cache_path).is_file():
        return chroot_native / custom_cache_path

    cache_path = gdk_cache_dir / "loaders.cache"
    if not (chroot_native / cache_path).is_file():
        raise RuntimeError(f"gdk pixbuf cache file not found: {cache_path}")

    pmb.chroot.root(["cp", cache_path, custom_cache_path])
    cmd: Sequence[PathString] = [
        "sed",
        "-i",
        "-e",
        f's@"{gdk_cache_dir}@"{chroot_native / gdk_cache_dir}@',
        custom_cache_path,
    ]
    pmb.chroot.root(cmd)
    return chroot_native / custom_cache_path


def command_qemu(
    config: Config,
    arch: Arch,
    cmdline_arg: str,
    qemu_audio: str | None,
    qemu_cpu: str | None,
    qemu_display: str,
    qemu_video: str,
    memory_size: int,
    port_ssh: int,
    use_host_qemu: bool,
    use_qemu_gl: bool,
    use_qemu_kvm: bool,
    use_qemu_tablet: bool,
    img_path: Path,
    img_path_2nd: Path | None = None,
) -> tuple[list[str | Path], Env]:
    """Generate the full qemu command with arguments to run postmarketOS"""
    chroot_native = Chroot.native()
    chroot_target = Chroot.buildroot(arch)

    ncpus = os.cpu_count()
    if not ncpus:
        logging.warning("Couldn't get cpu count, defaulting to 4")
        ncpus = 4

    # QEMU mach-virt's max CPU count is 8, limit it so it will work correctly
    # on systems with more than 8 CPUs
    if not arch.is_native() and ncpus > 8:
        ncpus = 8

    env: Env
    # It might be tempting to use PathString here, but I don't think it makes sense semantically as
    # this is not just a list of paths.
    command: list[str | Path]

    if use_host_qemu:
        qemu_bin = which_qemu(arch)
        env = {}
        command = [qemu_bin]
    else:
        env = {
            "QEMU_MODULE_DIR": chroot_native / "usr/lib/qemu",
            "GBM_DRIVERS_PATH": chroot_native / "usr/lib/xorg/modules/dri",
            "LIBGL_DRIVERS_PATH": chroot_native / "usr/lib/xorg/modules/dri",
        }

        if "gtk" in qemu_display:
            gdk_cache = create_gdk_loader_cache()
            env.update(
                {
                    "GTK_THEME": "Default",
                    "GDK_PIXBUF_MODULE_FILE": str(gdk_cache),
                    "XDG_DATA_DIRS": ":".join(
                        [
                            str(chroot_native / "usr/local/share"),
                            str(chroot_native / "usr/share"),
                        ]
                    ),
                }
            )

        command = []
        if Arch.native() in [Arch.aarch64, Arch.armv7]:
            # Workaround for QEMU failing on aarch64 asymmetric multiprocessor
            # arch (big/little architecture
            # https://en.wikipedia.org/wiki/ARM_big.LITTLE) see
            # https://bugs.linaro.org/show_bug.cgi?id=1443
            ncpus_bl = pmb.parse.cpuinfo.arm_big_little_first_group_ncpus()
            if ncpus_bl:
                ncpus = ncpus_bl
                logging.info(
                    "QEMU will run on big/little architecture on the"
                    f" first {ncpus} cores (from /proc/cpuinfo)"
                )
                command += [chroot_native / "lib" / f"ld-musl-{Arch.native()}.so.1"]
                command += [chroot_native / "usr/bin/taskset"]
                command += ["-c", "0-" + str(ncpus - 1)]

        command += [chroot_native / "lib" / f"ld-musl-{Arch.native()}.so.1"]
        command += [
            "--library-path="
            + ":".join(
                [
                    str(chroot_native / "lib"),
                    str(chroot_native / "usr/lib"),
                    str(chroot_native / "usr/lib/pulseaudio"),
                ]
            )
        ]
        command += [chroot_native / "usr/bin" / f"qemu-system-{arch.qemu_system()}"]
        command += ["-L", chroot_native / "usr/share/qemu/"]

    command += ["-nodefaults"]
    command += ["-smp", str(ncpus)]
    command += ["-m", str(memory_size)]

    command += ["-serial"]
    if config.qemu_redir_stdio:
        command += ["mon:stdio"]
    else:
        command += ["stdio"]

    command += ["-drive", f"file={img_path},format=raw,if=virtio"]
    if img_path_2nd:
        command += ["-drive", f"file={img_path_2nd}" + ",format=raw,if=virtio"]

    if use_qemu_tablet:
        command += ["-device", "virtio-tablet-pci"]
    else:
        command += ["-device", "virtio-mouse-pci"]
    command += ["-device", "virtio-keyboard-pci"]
    command += ["-netdev", f"user,id=net,hostfwd=tcp:127.0.0.1:{port_ssh}-:22"]
    command += ["-device", "virtio-net-pci,netdev=net"]

    # Check that arch is supported by pmb qemu
    if arch not in [
        Arch.x86_64,
        Arch.aarch64,
        Arch.riscv64,
        Arch.ppc64le,
        Arch.loongarch64,
        Arch.armv7,
    ]:
        raise RuntimeError(f"Architecture {arch} not supported by this command yet.")

    # Apply arch-specific qemu config
    match arch:
        case Arch.armv7:
            command += ["-M", "virt"]
            command += ["-cpu", "cortex-a15"]
        case Arch.aarch64:
            command += ["-M", "virt"]
            command += ["-cpu", "cortex-a57"]
        case Arch.riscv64:
            command += ["-M", "virt"]
        case Arch.ppc64le:
            command += ["-M", "pseries"]
        case Arch.loongarch64:
            command += ["-M", "virt"]
        case _:
            pass

    # Configure display
    if qemu_display != "none":
        match arch:
            case Arch.x86_64:
                command += ["-device", "virtio-vga-gl"]
            case Arch.aarch64 | Arch.armv7:
                command += ["-device", "virtio-gpu-gl"]
            case Arch.riscv64 | Arch.ppc64le | Arch.loongarch64:
                command += ["-device", "virtio-gpu-pci"]
            case _:
                # default to 2D-only backend
                command += ["-device", "virtio-gpu"]

    host_arch = Arch.native()
    edk2_chroot = chroot_native.path if arch == host_arch else chroot_target.path
    match arch:
        case Arch.aarch64:
            command += [
                "-drive",
                f"if=pflash,format=raw,readonly=on,file={edk2_chroot}/usr/share/AAVMF/AAVMF_CODE.fd",
            ]
        case Arch.x86_64:
            command += [
                "-drive",
                f"if=pflash,format=raw,readonly=on,file={edk2_chroot}/usr/share/OVMF/OVMF.fd",
            ]
        case Arch.loongarch64:
            command += [
                "-drive",
                f"if=pflash,format=raw,readonly=on,file={edk2_chroot}/usr/share/edk2/loongarch64/QEMU_EFI.fd",
            ]
        case Arch.ppc64le:
            # ppc64le uses SLOF included in QEMU
            pass
        case Arch.riscv64:
            command += [
                "-drive",
                f"if=pflash,format=raw,readonly=on,file={edk2_chroot}/usr/share/edk2/riscv/RISCV_VIRT_CODE.fd",
            ]
        case _:
            raise RuntimeError(f"Architecture {arch} not configured for EFI support.")

    # Kernel Virtual Machine (KVM) support
    native = pmb.parse.deviceinfo().arch.is_native()
    if use_qemu_kvm and native and os.path.exists("/dev/kvm"):
        command += ["-enable-kvm"]
        command += ["-cpu", "host"]
    else:
        logging.info("WARNING: QEMU is not using KVM and will run slower!")

    if qemu_cpu:
        command += ["-cpu", qemu_cpu]

    display = qemu_display
    if display != "none":
        display += ",show-cursor=on,gl=" + ("on" if use_qemu_gl else "off")

    # Separate -show-cursor option is deprecated. If your host qemu fails here,
    # it's old (#1995).
    command += ["-display", f"{display}"]

    # Audio support
    if qemu_audio:
        command += ["-audio", f"{qemu_audio},model=hda"]

    return (command, env)


def resize_image(img_size_new: str, img_path: Path) -> None:
    """
    Truncates an image to a specific size. The value must be larger than the
    current image size, and it must be specified in MiB or GiB units (powers of
    1024).

    :param img_size_new: new image size in M or G
    :param img_path: the path to the image
    """
    # Current image size in bytes
    img_size = os.path.getsize(img_path)

    # Make sure we have at least 1 integer followed by either M or G
    pattern = re.compile(r"^[0-9]+[M|G]$")
    if not pattern.match(img_size_new):
        raise RuntimeError("IMAGE_SIZE must be in [M]iB or [G]iB, e.g. 2048M or 2G")

    # Remove M or G and convert to bytes
    img_size_new_bytes = int(img_size_new[:-1]) * 1024 * 1024

    # Convert further for G
    if img_size_new[-1] == "G":
        img_size_new_bytes = img_size_new_bytes * 1024

    if img_size_new_bytes >= img_size:
        logging.info(f"Resize image to {img_size_new}: {img_path}")
        pmb.helpers.run.root(["truncate", "-s", img_size_new, img_path])
    else:
        # Convert to human-readable format
        # NOTE: We convert to M here, and not G, so that we don't have to
        # display a size like 1.25G, since decimal places are not allowed by
        # truncate.
        # We don't want users thinking they can use decimal numbers, and so in
        # this example, they would need to use a size greater then 1280M
        # instead.
        img_size_str = str(round(img_size / 1024 / 1024)) + "M"

        raise RuntimeError(f"IMAGE_SIZE must be {img_size_str} or greater")


def _sigterm_handler(number: int, stack_frame: FrameType | None) -> None:
    raise RuntimeError(
        "pmbootstrap was terminated by another process, and killed the QEMU VM it was running."
    )


def install_depends(arch: Arch) -> None:
    """Install any necessary qemu dependencies in native chroot"""
    depends = [
        "mesa-dri-gallium",
        "mesa-egl",
        "mesa-gl",
        "qemu",
        "qemu-audio-alsa",
        "qemu-audio-pa",
        "qemu-audio-sdl",
        "qemu-hw-display-virtio-gpu",
        "qemu-hw-display-virtio-gpu-gl",
        "qemu-hw-display-virtio-gpu-pci",
        "qemu-hw-display-virtio-gpu-pci-gl",
        "qemu-hw-display-virtio-vga",
        "qemu-hw-display-virtio-vga-gl",
        "qemu-system-" + arch.qemu_system(),
        "qemu-ui-gtk",
        "qemu-ui-opengl",
        "qemu-ui-sdl",
    ]

    chroot = Chroot.native()
    pmb.chroot.init(chroot)
    pmb.chroot.apk.install(depends, chroot)


def install_efi_image(arch: Arch) -> None:
    edk2_pkg = None

    # EDK2 builds are only available for the target architecture
    match arch:
        case Arch.aarch64:
            edk2_pkg = "aavmf"
        case Arch.x86_64:
            edk2_pkg = "ovmf"
        case Arch.loongarch64:
            edk2_pkg = "edk2-loongarch64"
        case Arch.ppc64le:
            # ppc64le uses SLOF included in QEMU instead of EFI
            return
        case Arch.riscv64:
            edk2_pkg = "ovmf"
        case _:
            raise RuntimeError(f"Architecture {arch} not configured for EFI support.")

    # If we're running natively, install it to the native chroot, otherwise we need to
    # install it in the target chroot
    target_chroot = Chroot.native() if arch == Arch.native() else Chroot.buildroot(arch)
    pmb.chroot.init(target_chroot)
    pmb.chroot.apk.install([edk2_pkg], target_chroot)


def qemu(
    cmdline_arg: str,
    qemu_audio: str | None,
    qemu_cpu: str | None,
    qemu_display: str,
    qemu_video: str,
    memory_size: int,
    image_size: str | None,
    second_storage: str | None,
    port_ssh: int,
    use_efi: bool,
    use_host_qemu: bool,
    use_qemu_gl: bool,
    use_qemu_kvm: bool,
    use_qemu_tablet: bool,
) -> None:
    """Run a postmarketOS image in qemu"""
    config = get_context().config
    device = config.device
    if use_efi:
        logging.warning(
            "WARNING: The --efi argument to 'pmbootstrap qemu' is deprecated. EFI boot is the default now and the argument no longer has any effect."
        )
    if not device.startswith("qemu-"):
        raise NonBugError(
            "'pmbootstrap qemu' can be only used with one of "
            "the QEMU device packages. Run 'pmbootstrap init' "
            "and select the 'qemu' vendor."
        )
    arch = pmb.parse.deviceinfo().arch

    # Make sure the rootfs image isn't mounted
    pmb.mount.umount_all(Chroot(ChrootType.IMAGE, "").path)
    pmb.install.losetup.detach_all()

    img_path = system_image(device)
    img_path_2nd = None
    if second_storage:
        img_path_2nd = create_second_storage(second_storage, device)

    if not use_host_qemu:
        install_depends(arch)
    install_efi_image(arch)

    logging.info("Running postmarketOS in QEMU VM (" + arch.qemu_system() + ")")

    qemu, env = command_qemu(
        config,
        arch,
        cmdline_arg,
        qemu_audio,
        qemu_cpu,
        qemu_display,
        qemu_video,
        memory_size,
        port_ssh,
        use_host_qemu,
        use_qemu_gl,
        use_qemu_kvm,
        use_qemu_tablet,
        img_path,
        img_path_2nd,
    )

    # Workaround: QEMU runs as local user and needs write permissions in the
    # rootfs, which is owned by root
    if not os.access(img_path, os.W_OK):
        pmb.helpers.run.root(["chmod", "666", img_path])

    # Resize the rootfs (or show hint)
    if image_size:
        resize_image(image_size, img_path)
    else:
        logging.info(
            "NOTE: Run 'pmbootstrap qemu --image-size 2G' to set"
            " the rootfs size when you run out of space!"
        )

    # SSH/serial/network hints
    logging.info("Connect to the VM:")
    logging.info(f"* (ssh) ssh -p {port_ssh} {config.user}@localhost")
    logging.info("* (serial) in this console (stdout/stdin)")

    if config.qemu_redir_stdio:
        logging.info(
            "NOTE: Ctrl+C is redirected to the VM! To disable this, "
            "run: pmbootstrap config qemu_redir_stdio False"
        )
        logging.info("NOTE: To quit QEMU with this option you can use Ctrl-A, X.")

    if config.ui == "none":
        logging.warning(
            "WARNING: With UI=none network doesn't work"
            " automatically: https://postmarketos.org/qemu-network"
        )

    # Run QEMU and kill it together with pmbootstrap
    process = None
    try:
        signal.signal(signal.SIGTERM, _sigterm_handler)
        process = pmb.helpers.run.user(qemu, output=RunOutputTypeDefault.TUI, env=env)
    except KeyboardInterrupt:
        # In addition to not showing a trace when pressing ^C, let user know
        # they can override this behavior:
        logging.info("Quitting because Ctrl+C detected.")
        logging.info("To override this behavior and have pmbootstrap send Ctrl+C to the VM, run:")
        logging.info("$ pmbootstrap config qemu_redir_stdio True")
    finally:
        if isinstance(process, subprocess.Popen):
            process.terminate()



# Copyright 2024 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.config.pmaports
import pmb.helpers.ui
from pmb.core import Config
from pmb.core.config import SystemdConfig
from pmb.meta import Cache


@Cache()
def is_systemd_selected(config: Config) -> bool:
    if "systemd" not in pmb.config.pmaports.read_config_repos():
        return False
    if pmb.helpers.ui.check_option(
        config.ui, "pmb:systemd-never", with_extra_repos="disabled", must_exist=False
    ):
        return False
    if config.systemd == SystemdConfig.ALWAYS:
        return True
    if config.systemd == SystemdConfig.NEVER:
        return False
    return pmb.helpers.ui.check_option(
        config.ui, "pmb:systemd", with_extra_repos="disabled", must_exist=False
    )


def systemd_selected_str(config: Config) -> tuple[str, str]:
    if "systemd" not in pmb.config.pmaports.read_config_repos():
        return "no", "not supported by pmaports branch"
    if pmb.helpers.ui.check_option(config.ui, "pmb:systemd-never", must_exist=False):
        return "no", "not supported by selected UI"
    if config.systemd == SystemdConfig.ALWAYS:
        return "yes", "'always' selected in 'pmbootstrap init'"
    if config.systemd == SystemdConfig.NEVER:
        return "no", "'never' selected in 'pmbootstrap init'"
    if pmb.helpers.ui.check_option(config.ui, "pmb:systemd", must_exist=False):
        return "yes", "default for selected UI"
    return "no", "default for selected UI"



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import sys

import pmb.config
import pmb.helpers.git
from pmb.core.context import CommandTimeout, Context, TimeoutReason
from pmb.core.pkgrepo import pkgrepo_default_path
from pmb.types import PmbArgs

"""This file constructs the args variable, which is passed to almost all
   functions in the pmbootstrap code base. Here's a listing of the kind of
   information it stores.

    1. Argparse
       Variables directly from command line argument parsing (see
       pmb/parse/arguments.py, the "dest" parameter of the add_argument()
       calls defines where it is stored in args).

       Examples:
       args.action ("zap", "chroot", "build" etc.)
       args.as_root (True when --as-root is passed)
       ...

    2. Argparse merged with others
       Variables from the user's config file (~/.config/pmbootstrap_v3.cfg) that
       can be overridden from the command line (pmb/parse/arguments.py) and
       fall back to the defaults defined in pmb/config/__init__.py (see
       "defaults = {..."). The user's config file gets generated interactively
        with "pmbootstrap init".

       Examples:
       args.aports ("$WORK/cache_git/pmaports", override with --aports)
       args.device ("samsung-i9100", "qemu-amd64" etc.)
       get_context().config.work ("/home/user/.local/var/pmbootstrap", override with --work)

    3. Parsed configs
       Similar to the cache above, specific config files get parsed and added
       to args, so they can get accessed quickly (without parsing the configs
       over and over). These configs are not only used in one specific
       location, so having a short name for them increases readability of the
       code as well.

       Examples:
       deviceinfo (e.g. {"name": "Mydevice", "arch": "armhf", ...})
"""


def init(args: PmbArgs) -> PmbArgs:
    args_ = PmbArgs()
    # Basic initialization
    # print(json.dumps(args.__dict__))
    # sys.exit(0)
    config = pmb.config.load(args.config)

    if args.aports:
        for pmaports_dir in args.aports:
            if pmaports_dir.exists():
                continue
            raise ValueError(
                f"pmaports path (specified with --aports) does not exist: {pmaports_dir}"
            )

    # Override config at runtime with command line arguments
    for key in vars(config):
        if key.startswith("_") or key == "user":
            continue
        value = getattr(args, key, None)
        if value:
            setattr(config, key, value)

        # Deny accessing the attribute via args
        if hasattr(args, key):
            delattr(args, key)

    running_in_ci = os.environ.get("CI")

    # Configure runtime context
    context = Context(config)
    # Only enable the timeout by default in CI.
    if running_in_ci and args.timeout is None:
        context.command_timeout = CommandTimeout(length=900, reason=TimeoutReason.CI_DETECTED)
    elif args.timeout is not None:
        context.command_timeout = CommandTimeout(
            length=args.timeout, reason=TimeoutReason.TIMEOUT_ARG
        )
    context.details_to_stdout = args.details_to_stdout
    context.quiet = args.quiet
    context.offline = args.offline
    context.command = args.action
    context.cross = args.cross
    context.ccache = args.ccache
    context.assume_yes = getattr(args, "assume_yes", False)
    context.force = getattr(args, "force", False)

    # Initialize context
    pmb.core.context.set_context(context)

    # Initialize logs (we could raise errors below)
    pmb.helpers.logging.init(context.log, args.verbose, context.details_to_stdout)
    pmb.helpers.logging.debug(f"Pmbootstrap v{pmb.__version__} (Python {sys.version})")

    if context.command_timeout:
        pmb.helpers.logging.info(
            f"NOTE: Timeout enabled, commands will be killed after {context.command_timeout.length} seconds of not writing any output. Reason: {context.command_timeout.reason}"
        )

    # Initialization code which may raise errors
    if args.action not in [
        "init",
        "checksum",
        "config",
        "bootimg_analyze",
        "log",
        "pull",
        "shutdown",
        "zap",
    ]:
        pmb.config.pmaports.read_config()
        pmb.helpers.git.parse_channels_cfg(pkgrepo_default_path())

    # Remove attributes from args so they don't get used by mistake
    if hasattr(args, "timeout"):
        del args.timeout
    del args.details_to_stdout
    del args.log
    del args.quiet
    del args.offline
    if hasattr(args, "force"):
        del args.force
    if hasattr(args, "device"):
        del args.device

    # Copy all properties from args to out that don't start with underscores
    for key, value in vars(args).items():
        if not key.startswith("_") and key != "from_argparse":
            setattr(args_, key, value)

    return args_



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Functions that work with binary package repos.

See also:
- pmb/helpers/pmaports.py (work with pmaports)
- pmb/helpers/package.py (work with both)
"""

import hashlib
import os
from pathlib import Path
from typing import Literal

import pmb.config.pmaports
import pmb.helpers.http
import pmb.helpers.run
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.core.pkgrepo import pkgrepo_names
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError
from pmb.meta import Cache


def apkindex_hash(url: str, length: int = 8) -> Path:
    r"""
    Generate the hash that APK adds to the APKINDEX and apk packages in its apk cache folder.

    It is the "12345678" part in this example:
    "APKINDEX.12345678.tar.gz".

    :param length: The length of the hash in the output file.

    See also: official implementation in apk-tools:
    <https://git.alpinelinux.org/cgit/apk-tools/>

    blob.c: apk_blob_push_hexdump(), "const char \\*xd"
    apk_defines.h: APK_CACHE_CSUM_BYTES
    database.c: apk_repo_format_cache_index()
    """
    binary = hashlib.sha1(url.encode("utf-8")).digest()
    xd = "0123456789abcdefghijklmnopqrstuvwxyz"
    csum_bytes = int(length / 2)

    ret = ""
    for i in range(csum_bytes):
        ret += xd[(binary[i] >> 4) & 0xF]
        ret += xd[binary[i] & 0xF]

    return Path(f"APKINDEX.{ret}.tar.gz")


# FIXME: make config.mirrors a normal dict
# mypy: disable-error-code="literal-required"
@Cache("user_repository", "mirrors_exclude")
def urls(
    user_repository: Path | None = None, mirrors_exclude: list[str] | Literal[True] = []
) -> list[str]:
    """
    Get a list of repository URLs, as they are in /etc/apk/repositories.

    :param user_repository: add /mnt/pmbootstrap/packages
    :param mirrors_exclude: mirrors to exclude (see pmb.core.config.Mirrors) or true to exclude
                            all mirrors and only return the local repos
    :returns: list of mirror strings, like ["/mnt/pmbootstrap/packages",
                                            "http://...", ...]
    """
    ret: list[str] = []
    config = get_context().config

    # Get mirrordirs from channels.cfg (postmarketOS mirrordir is the same as
    # the pmaports branch of the channel, no need to make it more complicated)
    channel_cfg = pmb.config.pmaports.read_config_channel()
    mirrordir_pmos = channel_cfg["branch_pmaports"]
    mirrordir_alpine = channel_cfg["mirrordir_alpine"]

    # Local user repository (for packages compiled with pmbootstrap)
    if user_repository:
        ret.extend(str(user_repository / channel) for channel in pmb.config.pmaports.all_channels())

    if mirrors_exclude is True:
        return ret

    # Don't add the systemd mirror if systemd is disabled
    if not pmb.config.is_systemd_selected(config):
        mirrors_exclude.append("systemd")

    # ["pmaports", "systemd", "alpine", "plasma-nightly"]
    for repo in [*pkgrepo_names(), "alpine"]:
        if repo in mirrors_exclude:
            continue

        # Allow adding a custom mirror in front of the real mirror. This is used
        # in bpo to build with a WIP repository in addition to the final
        # repository.
        for suffix in ["_custom", ""]:
            mirror = config.mirrors[f"{repo}{suffix}"]

            # During bootstrap / bpo testing we run without a pmOS binary repo
            if mirror.lower() == "none":
                if suffix != "_custom":
                    logging.warn_once(
                        f"NOTE: Skipping mirrors.{repo} for /etc/apk/repositories (is configured"
                        ' as "none")'
                    )
                continue

            mirrordirs = []
            if repo == "alpine":
                # FIXME: This is a bit of a mess
                mirrordirs = [f"{mirrordir_alpine}/main", f"{mirrordir_alpine}/community"]
                if mirrordir_alpine == "edge":
                    mirrordirs.append(f"{mirrordir_alpine}/testing")
            else:
                mirrordirs = [mirrordir_pmos]

            for mirrordir in mirrordirs:
                url = os.path.join(mirror, mirrordir)
                if url not in ret:
                    ret.append(url)

    return ret


def apkindex_files(
    arch: Arch | None = None, user_repository: bool = True, exclude_mirrors: list[str] = []
) -> list[Path]:
    """
    Get a list of outside paths to all resolved APKINDEX.tar.gz files for a specific arch.

    :param arch: defaults to native
    :param user_repository: add path to index of locally built packages
    :param exclude_mirrors: list of mirrors to exclude (e.g. ["alpine", "pmaports"])
    :returns: list of absolute APKINDEX.tar.gz file paths
    """
    if not arch:
        arch = Arch.native()

    ret: list[Path] = []
    # Local user repository (for packages compiled with pmbootstrap)
    if user_repository:
        ret.extend(
            get_context().config.work / "packages" / channel / arch / "APKINDEX.tar.gz"
            for channel in pmb.config.pmaports.all_channels()
        )

    # Resolve the APKINDEX.$HASH.tar.gz files
    ret.extend(
        file
        for url in urls(False, exclude_mirrors)
        if (file := get_context().config.work / f"cache_apk_{arch}" / apkindex_hash(url)).exists()
    )

    return ret


@Cache("arch", force=False)
def update(arch: Arch | None = None, force: bool = False, existing_only: bool = False) -> bool:
    """
    Download the APKINDEX files for all URLs depending on the architectures.

    :param arch: * one Alpine architecture name ("x86_64", "armhf", ...)
                 * None for all architectures
    :param force: even update when the APKINDEX file is fairly recent
    :param existing_only: only update the APKINDEX files that already exist,
                          this is used by "pmbootstrap update"

    :returns: True when files have been downloaded, False otherwise
    """
    # Skip in offline mode, only show once
    if get_context().offline:
        logging.warn_once("NOTE: skipping package index update (offline mode)")
        return False

    # Architectures and retention time
    supported_binary = Arch.supported_binary()
    architectures = [arch] if arch else supported_binary
    retention_hours = pmb.config.apkindex_retention_time
    retention_seconds = retention_hours * 3600

    # Find outdated APKINDEX files. Formats:
    # outdated: {URL: apkindex_path, ... }
    # outdated_arches: ["armhf", "x86_64", ... ]
    outdated = {}
    outdated_arches: list[Arch] = []
    for url in urls(False):
        for architecture in architectures:
            # APKINDEX file name from the URL
            url_full = f"{url}/{architecture}/APKINDEX.tar.gz"
            cache_apk_outside = get_context().config.work / f"cache_apk_{architecture}"
            apkindex = cache_apk_outside / f"{apkindex_hash(url)}"

            # Find update reason, possibly skip non-existing or known 404 files
            reason = None
            if not os.path.exists(apkindex):
                if existing_only:
                    continue
                reason = "file does not exist yet"
            elif force:
                reason = "forced update"
            elif pmb.helpers.file.is_older_than(apkindex, retention_seconds):
                reason = "older than " + str(retention_hours) + "h"
            if not reason:
                continue

            # Update outdated and outdated_arches
            logging.debug("APKINDEX outdated (" + reason + "): " + url_full)
            outdated[url_full] = apkindex
            if architecture not in outdated_arches:
                outdated_arches.append(architecture)

    # Bail out or show log message
    if not len(outdated):
        return False
    logging.info(
        "Update package index for "
        + ", ".join([str(a) for a in outdated_arches])
        + " ("
        + str(len(outdated))
        + " file(s))"
    )

    # Download and move to right location
    missing_ignored = False
    for i, (url, target) in enumerate(outdated.items()):
        pmb.helpers.cli.progress_print(i / len(outdated))
        temp = pmb.helpers.http.download(url, "APKINDEX", False, logging.DEBUG, True, True)
        if not temp:
            if (
                os.environ.get("PMB_APK_FORCE_MISSING_REPOSITORIES") == "1"
                or arch not in supported_binary
            ):
                missing_ignored = True
                continue
            else:
                logging.info("NOTE: check the [mirrors] section in 'pmbootstrap config'")
                raise NonBugError("getting APKINDEX from binary package mirror failed!")
        target_folder = os.path.dirname(target)
        if not os.path.exists(target_folder):
            pmb.helpers.run.root(["mkdir", "-p", target_folder])
        pmb.helpers.run.root(["cp", temp, target])
    pmb.helpers.cli.progress_flush()

    if missing_ignored:
        logging.warn_once(
            "NOTE: ignoring missing APKINDEX due to PMB_APK_FORCE_MISSING_REPOSITORIES=1 or architecture without binary repo (fine during bootstrap)"
        )

    return True


def alpine_apkindex_path(repo: str = "main", arch: Arch | None = None) -> Path:
    """
    Get the path to a specific Alpine APKINDEX file on disk and download it if necessary.

    :param repo: Alpine repository name (e.g. "main")
    :param arch: Alpine architecture (e.g. "armhf"), defaults to native arch.
    :returns: full path to the APKINDEX file
    """
    # Repo sanity check
    if repo not in ["main", "community", "testing", "non-free"]:
        raise RuntimeError(f"Invalid Alpine repository: {repo}")

    # Download the file
    arch = arch or Arch.native()
    update(arch)

    # Find it on disk
    channel_cfg = pmb.config.pmaports.read_config_channel()
    repo_link = f"{get_context().config.mirrors['alpine']}{channel_cfg['mirrordir_alpine']}/{repo}"
    cache_folder = get_context().config.work / (f"cache_apk_{arch}")
    return cache_folder / apkindex_hash(repo_link)



# Copyright 2024 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.config
import pmb.config.other
import pmb.helpers.git
from pmb.core import Config
from pmb.core.context import get_context


def print_status_line(key: str, value: str) -> None:
    styles = pmb.config.styles
    key = f"{styles['GREEN']}{key}{styles['END']}:"
    padding = 17

    print(f"{key.ljust(padding)} {value}")


def print_channel(config: Config) -> None:
    pmaports_cfg = pmb.config.pmaports.read_config()
    channel = pmaports_cfg["channel"]

    # Get branch name (if on branch) or current commit
    path = pmb.helpers.git.get_path("pmaports")
    ref = pmb.helpers.git.rev_parse(path, extra_args=["--abbrev-ref"], silent=True)
    if ref == "HEAD":
        ref = pmb.helpers.git.rev_parse(path, silent=True)[0:8]

    if not pmb.helpers.git.clean_worktree(path, silent=True):
        ref += ", dirty"

    value = f"{channel} (pmaports: {ref})"
    print_status_line("Channel", value)


def print_device(config: Config) -> None:
    kernel = ""
    if pmb.parse._apkbuild.kernels(config.device):
        kernel = f", kernel: {config.kernel}"

    value = f"{config.device} ({pmb.parse.deviceinfo().arch}{kernel})"
    print_status_line("Device", value)


def print_ui(config: Config) -> None:
    print_status_line("UI", config.ui)


def print_systemd(config: Config) -> None:
    yesno, reason = pmb.config.other.systemd_selected_str(config)
    print_status_line("systemd", f"{yesno} ({reason})")


def print_status() -> None:
    """
    :param details: if True, print each passing check instead of a summary
    :returns: True if all checks passed, False otherwise
    """
    config = get_context().config
    print_channel(config)
    print_device(config)
    print_ui(config)
    print_systemd(config)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
from enum import Enum
from typing import Any

import pmb.helpers.file
import pmb.helpers.pmaports
import pmb.helpers.repo
import pmb.parse
import pmb.parse.apkindex
from pmb.core.apkindex_block import ApkindexBlock
from pmb.core.arch import Arch
from pmb.helpers import logging


class BumpType(Enum):
    PKGREL = "pkgrel"
    PKGVER = "pkgver"


def package(
    pkgname: str, reason: str = "", dry: bool = False, bump_type: BumpType = BumpType.PKGREL
) -> None:
    """
    Increase the pkgrel or pkgver in the APKBUILD of a specific package.

    :param pkgname: name of the package
    :param reason: string to display as reason why it was increased
    :param dry: don't modify the APKBUILD, just print the message
    :param bump_type: whether to bump pkgrel or pkgver
    """
    # Current and new pkgrel or pkgver
    path = pmb.helpers.pmaports.find(pkgname) / "APKBUILD"
    apkbuild = pmb.parse.apkbuild(path)
    version = int(apkbuild[bump_type.value])
    version_new = version + 1

    # Display the message, bail out in dry mode
    logging.info(
        "Increase '"
        + pkgname
        + f"' {bump_type.value} ("
        + str(version)
        + " -> "
        + str(version_new)
        + ")"
        + reason
    )
    if dry:
        return

    # Increase
    old = f"\n{bump_type.value}=" + str(version) + "\n"
    new = f"\n{bump_type.value}=" + str(version_new) + "\n"
    pmb.helpers.file.replace(path, old, new)

    if bump_type == BumpType.PKGVER:
        pkgrel = int(apkbuild["pkgrel"])
        # Set pkgrel to 0 if we bump pkgver
        pmb.helpers.file.replace(path, f"pkgrel={pkgrel}", "pkgrel=0")

    # Verify
    pmb.parse.apkbuild.cache_clear()
    apkbuild = pmb.parse.apkbuild(path)
    if int(apkbuild[bump_type.value]) != version_new:
        raise RuntimeError(
            f"Failed to bump {bump_type.value} for package '{pkgname}'."
            " Make sure that there's a line with exactly the"
            f" string '{old.strip()}' and nothing else in: {path}"
        )


def auto_apkindex_package(
    arch: Arch, aport: dict[str, Any], apk: ApkindexBlock, dry: bool = False
) -> bool:
    """
    Bump the pkgrel of a specific package if it is outdated in the given APKINDEX.

    :param arch: the architecture, e.g. "armhf"
    :param aport: parsed APKBUILD of the binary package's origin:
                  {"pkgname": ..., "pkgver": ..., "pkgrel": ..., ...}
    :param apk: information about the binary package from the APKINDEX:
                {"version": ..., "depends": [...], ...}
    :param dry: don't modify the APKBUILD, just print the message
    :returns: True when there was an APKBUILD that needed to be changed.
    """
    version_aport = aport["pkgver"] + "-r" + aport["pkgrel"]
    version_apk = apk.version
    pkgname = aport["pkgname"]

    # Skip when aport version != binary package version
    compare = pmb.parse.version.compare(version_aport, version_apk)
    if compare == -1:
        logging.warning(
            f"{pkgname}: skipping, because the aport version {version_aport} is lower"
            f" than the binary version {version_apk}"
        )
        return False
    if compare == 1:
        logging.verbose(
            f"{pkgname}: skipping, because the aport version {version_aport} is higher"
            f" than the binary version {version_apk}"
        )
        return False

    # Find missing depends
    logging.verbose("{}: checking depends: {}".format(pkgname, ", ".join(apk.depends)))
    missing = []
    for depend in apk.depends:
        if depend.startswith("!"):
            # Ignore conflict-dependencies
            continue

        providers = pmb.parse.apkindex.providers(depend, arch, must_exist=False)
        if providers == {} and (
            # We're only interested in missing depends starting with "so:"
            # (which means dynamic libraries that the package was linked
            # against) and packages for which no aport exists.
            depend.startswith("so:") or not pmb.helpers.pmaports.find_optional(depend)
        ):
            missing.append(depend)

    # Increase pkgrel
    if len(missing):
        package(pkgname, reason=", missing depend(s): " + ", ".join(missing), dry=dry)
        return True

    return False


def auto(dry: bool = False) -> list[str]:
    """:returns: list of aport names, where the pkgrel needed to be changed"""
    ret = []
    for arch in Arch.supported():
        paths = pmb.helpers.repo.apkindex_files(arch, exclude_mirrors=["alpine"])
        for path in paths:
            logging.info(f"scan {path}")
            index = pmb.parse.apkindex.parse(path, False)
            for pkgname, apk in index.items():
                origin = apk.origin
                # Only increase once!
                if origin in ret:
                    logging.verbose(f"{pkgname}: origin '{origin}' found again")
                    continue

                if origin is None:
                    logging.warning(f"{pkgname}: skipping, is a virtual package")
                    continue

                aport_path = pmb.helpers.pmaports.find_optional(origin)
                if not aport_path:
                    logging.warning(f"{pkgname}: origin '{origin}' aport not found")
                    continue
                aport = pmb.parse.apkbuild(aport_path)
                if auto_apkindex_package(arch, aport, apk, dry):
                    ret.append(pkgname)
    return ret



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Functions that work with both pmaports and binary package repos.

See also:

    - pmb/helpers/pmaports.py (work with pmaports)

    - pmb/helpers/repo.py (work with binary package repos)
"""

from typing import overload

import pmb.helpers.pmaports
import pmb.helpers.repo
import pmb.parse.version
from pmb.core.arch import Arch
from pmb.core.package_metadata import PackageMetadata
from pmb.helpers import logging
from pmb.meta import Cache


def remove_operators(package: str) -> str:
    for operator in [">", ">=", "<=", "=", "<", "~"]:
        if operator in package:
            package = package.split(operator)[0]
            break
    return package


def check_version_constraints(pkgname_with_op: str, version: str) -> bool:
    # Operators and matching return values from pmb.parse.version.compare()
    operators = {
        ">=": [0, 1],
        ">": [1],
        "<=": [-1, 0],
        "<": [-1],
    }

    for op, valid_results in operators.items():
        if op not in pkgname_with_op:
            continue
        ver_req = pkgname_with_op.split(op, 1)[1]
        result = pmb.parse.version.compare(version, ver_req)
        logging.verbose(
            f"check_version_constraints: op:{op}, version:{version}, ver_req:{ver_req}, result:{result}"
        )
        if result in valid_results:
            logging.debug(f"{pkgname_with_op}: matches package")
            return True
        else:
            logging.debug(f"{pkgname_with_op}: does not match package")
            return False

    logging.debug(f"check_version_constraints: ignoring {pkgname_with_op}")
    return True


@overload
def get(pkgname: str, arch: Arch, replace_subpkgnames: bool = ...) -> PackageMetadata: ...


@overload
def get(
    pkgname: str, arch: Arch, replace_subpkgnames: bool = ..., must_exist: bool = ...
) -> PackageMetadata | None: ...


@overload
def get(
    pkgname: str,
    arch: Arch,
    replace_subpkgnames: bool = ...,
    must_exist: bool = ...,
    try_other_arches: bool = ...,
) -> PackageMetadata | None: ...


@Cache("pkgname", "arch", "replace_subpkgnames", "try_other_arches")
def get(
    pkgname: str,
    arch: Arch,
    replace_subpkgnames: bool = False,
    must_exist: bool = True,
    try_other_arches: bool = True,
) -> PackageMetadata | None:
    """
    Find a package in pmaports, and as fallback in the APKINDEXes of the binary packages.

    :param pkgname: package name (e.g. "hello-world")
    :param arch: preferred architecture of the binary package.
        When it can't be found for this arch, we'll still look for another arch to see whether the
        package exists at all. So make sure to check the returned arch against what you wanted
        with check_arch(). Example: "armhf"
    :param replace_subpkgnames: replace all subpkgnames with their main pkgnames in the depends
        (see #1733)
    :param must_exist: raise an exception, if not found
    :param try_other_arches: set to False to not attempt to find other arches

    :returns: * data from the parsed APKBUILD or APKINDEX in the following format:
                    {"arch": ["noarch"], "depends": ["busybox-extras", "lddtree", ...],
                    "pkgname": "postmarketos-mkinitfs", "provides": ["mkinitfs=0..1"],
                    "version": "0.0.4-r10"}

        * None if the package was not found
    """
    # Find in pmaports
    ret: PackageMetadata | None = None
    pmaport = pmb.helpers.pmaports.get(pkgname, False)
    if pmaport:
        ret = PackageMetadata.from_pmaport(pmaport)

    # Find in APKINDEX (given arch)
    if not ret or not pmb.helpers.pmaports.check_arches(ret.arch, arch):
        pmb.helpers.repo.update(arch)
        ret_repo = pmb.parse.apkindex.package(pkgname, arch, False)

        # Save as result if there was no pmaport, or if the pmaport can not be
        # built for the given arch, but there is a binary package for that arch
        # (e.g. temp/mesa can't be built for x86_64, but Alpine has it)
        if ret_repo and (not ret or ret_repo.arch == arch):
            ret = PackageMetadata.from_apkindex_block(ret_repo)

    # Find in APKINDEX (other arches)
    if not ret and try_other_arches:
        pmb.helpers.repo.update()
        for arch_i in Arch.supported():
            if arch_i != arch:
                apkindex_block = pmb.parse.apkindex.package(pkgname, arch_i, False)
                if apkindex_block is not None:
                    ret = PackageMetadata.from_apkindex_block(apkindex_block)
            if ret:
                break

    # Replace subpkgnames if desired
    if replace_subpkgnames and ret:
        depends_new = []
        for depend in ret.depends:
            depend_data = get(depend, arch, must_exist=False, try_other_arches=try_other_arches)
            if not depend_data:
                logging.warning(f"WARNING: {pkgname}: failed to resolve dependency '{depend}'")
                # Can't replace potential subpkgname
                if depend not in depends_new:
                    depends_new += [depend]
                continue
            depend_pkgname = depend_data.pkgname
            if depend_pkgname not in depends_new:
                depends_new += [depend_pkgname]
        ret.depends = depends_new

    # Save to cache and return
    if ret:
        return ret

    # Could not find the package
    if not must_exist:
        return None
    raise RuntimeError(f"Package '{pkgname}': Could not find it in pmaports or any APKINDEX!")


@Cache("pkgname", "arch")
def depends_recurse(pkgname: str, arch: Arch) -> list[str]:
    """
    Recursively resolve all of the package's dependencies.

    :param pkgname: name of the package (e.g. "device-samsung-i9100")
    :param arch: preferred architecture for binary packages
    :returns: a list of pkgname_start and all its dependencies, e.g:
        ["busybox-static-armhf", "device-samsung-i9100",
        "linux-samsung-i9100", ...]
    """
    # Build ret (by iterating over the queue)
    queue = {pkgname}
    ret: set[str] = set()
    alternative_names: set[str] = set()
    while len(queue):
        pkgname = queue.pop()
        package = get(pkgname, arch)

        # Add its depends to the queue
        for depend in package.depends:
            if depend not in ret and depend not in alternative_names:
                queue.add(depend)

        # Add the pkgname (not possible subpkgname) to ret
        ret.add(package.pkgname)
        alternative_names.update(package.provides)
        alternative_names.add(pkgname)

    return sorted(ret)


def check_arch(pkgname: str, arch: Arch, binary: bool = True) -> bool:
    """
    Check if a package be built for a certain architecture, or is there a binary package for it.

    :param pkgname: name of the package
    :param arch: architecture to check against
    :param binary: set to False to only look at the pmaports, not at binary
        packages

    :returns: True when the package can be built, or there is a binary package, False otherwise
    """
    if binary:
        arches = get(pkgname, arch).arch
    else:
        arches = pmb.helpers.pmaports.get(pkgname, must_exist=True)["arch"]
    return pmb.helpers.pmaports.check_arches(arches, arch)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
from pmb.install._install import (
    get_kernel_package as get_kernel_package,
    install as install,
)
from pmb.install.format import (
    format as format,
    get_root_filesystem as get_root_filesystem,
)
from pmb.install.partition import (
    partition as partition,
    partition_cgpt as partition_cgpt,
    partition_prep as partition_prep,
    partitions_mount as partitions_mount,
)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import tempfile

import pmb.chroot
from pmb.core import Chroot
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.devices import get_device_category_by_name
from pmb.types import PartitionLayout, PathString, RunOutputTypeDefault


def install_fsprogs(filesystem: str) -> None:
    """Install the package required to format a specific filesystem."""
    fsprogs = pmb.config.filesystems.get(filesystem)
    if not fsprogs:
        raise RuntimeError(f"Unsupported filesystem: {filesystem}")
    pmb.chroot.apk.install([fsprogs], Chroot.native())


def format_partition_with_filesystem(
    device: str, label: str, filesystem: str, is_disk: bool
) -> None:
    """
    :param device: partition to format with filesystem
    :param label: label to apply to filesystem
    :param filesystem: filesystem type to format partition with
    :param is_disk: whether we are formatting on physical disk
    """
    # Install binaries for formatting the selected filesystem
    install_fsprogs(filesystem)

    # Create formatting command
    if filesystem == "ext4":
        # ext4 uses more complex logic because it is used by downstream kernels
        device_category = get_device_category_by_name(get_context().config.device)
        if device_category.allows_downstream_ports():
            # Some downstream kernels don't support metadata_csum (#1364).
            # When changing the options of mkfs.ext4, also change them in the
            # recovery zip code (see 'grep -r mkfs\.ext4')!
            category_opts = ["-O", "^metadata_csum"]
        else:
            category_opts = []
        mkfs_args = ["mkfs.ext4", *category_opts, "-F", "-q", "-L", label]
        if not is_disk:
            # pmb#2568: tell mkfs.ext4 to make a filesystem with enough
            # indoes that we don't run into "out of space" errors
            mkfs_args = [*mkfs_args, "-i", "16384"]
    elif filesystem == "fat16":
        mkfs_args = ["mkfs.fat", "-F", "16", "-n", label]
    elif filesystem == "fat32":
        mkfs_args = ["mkfs.fat", "-F", "32", "-n", label]
    elif filesystem == "ext2":
        mkfs_args = ["mkfs.ext2", "-F", "-q", "-L", label]
    elif filesystem == "f2fs":
        mkfs_args = ["mkfs.f2fs", "-O extra_attr,inode_checksum,sb_checksum", "-f", "-l", label]
    elif filesystem == "btrfs":
        mkfs_args = ["mkfs.btrfs", "-f", "-q", "-L", label]
    elif filesystem == "xfs":
        # mkfs.xfs requires specifying the sector size
        mkfs_args = ["mkfs.xfs", "-f", "-q"]
        sector_size = pmb.parse.deviceinfo().rootfs_image_sector_size
        if sector_size is None or sector_size in {"", "512"}:
            mkfs_args += ["-s", "size=512"]
        else:
            mkfs_args += [
                "-b",
                "size=" + str(int(sector_size) * 2),
                "-s",
                "size=" + str(sector_size),
            ]
        mkfs_args += ["-L", label]
    else:
        raise RuntimeError("Filesystem " + filesystem + " is not supported!")

    # Format the partition with the selected filesystem
    logging.info(f"(native) format {device} ({label}, {filesystem})")
    pmb.chroot.root([*mkfs_args, device])


def format_and_mount_boot(device: str, boot_label: str) -> None:
    """
    :param device: boot partition on install block device (e.g. /dev/installp1)
    :param boot_label: label of the root partition (e.g. "pmOS_boot")
    """
    mountpoint = "/mnt/install/boot"
    filesystem = pmb.parse.deviceinfo().boot_filesystem or "ext2"
    format_partition_with_filesystem(device, boot_label, filesystem, False)

    pmb.chroot.root(["mkdir", "-p", mountpoint])
    pmb.chroot.root(["mount", device, mountpoint])


def format_luks_root(device: str, cipher: str, iter_time: str) -> None:
    """:param device: root partition on install block device (e.g. /dev/installp2)"""
    mountpoint = "/dev/mapper/pm_crypt"

    logging.info(f"(native) format {device} (root, luks), mount to {mountpoint}")
    logging.info(" *** TYPE IN THE FULL DISK ENCRYPTION PASSWORD (TWICE!) ***")

    # Avoid cryptsetup warning about missing locking directory
    pmb.chroot.root(["mkdir", "-p", "/run/cryptsetup"])

    format_cmd = [
        "cryptsetup",
        "luksFormat",
        "-q",
        "--cipher",
        cipher,
        "--iter-time",
        iter_time,
        "--use-random",
        device,
    ]
    open_cmd = ["cryptsetup", "luksOpen"]

    path_outside = None
    fde_key = os.environ.get("PMB_FDE_PASSWORD", None)
    if fde_key:
        # Write passphrase to a temp file, to avoid printing it in any log
        path = tempfile.mktemp(dir="/tmp")
        path_outside = Chroot.native() / path
        path_outside.write_text(fde_key, encoding="utf-8")
        format_cmd += [str(path)]
        open_cmd += ["--key-file", str(path)]

    try:
        pmb.chroot.root(format_cmd, output=RunOutputTypeDefault.INTERACTIVE)
        pmb.chroot.root([*open_cmd, device, "pm_crypt"], output=RunOutputTypeDefault.INTERACTIVE)
    finally:
        if path_outside:
            path_outside.unlink()

    if not (Chroot.native() / mountpoint).exists():
        raise RuntimeError("Failed to open cryptdevice!")


def get_root_filesystem(filesystem: str | None) -> str:
    ret = filesystem or pmb.parse.deviceinfo().root_filesystem or "ext4"
    pmaports_cfg = pmb.config.pmaports.read_config()

    supported = pmaports_cfg.get("supported_root_filesystems", "ext4")
    supported_list = supported.split(",")

    if ret not in supported_list:
        raise ValueError(
            f"Root filesystem {ret} is not supported by your"
            " currently checked out pmaports branch. Update your"
            " branch ('pmbootstrap pull'), change it"
            " ('pmbootstrap init'), or select one of these"
            f" filesystems: {', '.join(supported_list)}"
        )
    return ret


def prepare_btrfs_subvolumes(device: str, mountpoint: str) -> None:
    """
    Create separate subvolumes if root filesystem is btrfs.
    This lets us do snapshots and rollbacks of relevant parts
    of the filesystem.
    /var contains logs, VMs, containers, flatpaks; and shouldn't roll back,
    /root is root's home directory and shouldn't roll back,
    /tmp has temporary files, snapshotting them is unnecessary,
    /srv contains data for web and FTP servers, and shouldn't roll back,
    /snapshots should be a separate subvol so that changing the root subvol
    doesn't affect snapshots
    """
    subvolume_list = ["@", "@home", "@root", "@snapshots", "@srv", "@tmp", "@var"]

    for subvolume in subvolume_list:
        pmb.chroot.root(["btrfs", "subvol", "create", f"{mountpoint}/{subvolume}"])

    # Set the default root subvolume to be separate from top level btrfs
    # subvol. This lets us easily swap out current root subvol with an
    # earlier snapshot.
    pmb.chroot.root(["btrfs", "subvol", "set-default", f"{mountpoint}/@"])

    # Make directories to mount subvols onto
    pmb.chroot.root(["umount", mountpoint])
    pmb.chroot.root(["mount", device, mountpoint])
    pmb.chroot.root(
        [
            "mkdir",
            f"{mountpoint}/home",
            f"{mountpoint}/root",
            f"{mountpoint}/.snapshots",
            f"{mountpoint}/srv",
            f"{mountpoint}/var",
        ]
    )

    # snapshots contain sensitive information,
    # and should only be readable by root.
    pmb.chroot.root(["chmod", "700", f"{mountpoint}/root"])
    pmb.chroot.root(["chmod", "700", f"{mountpoint}/.snapshots"])

    # Mount subvols
    pmb.chroot.root(["mount", "-o", "subvol=@var", device, f"{mountpoint}/var"])
    pmb.chroot.root(["mount", "-o", "subvol=@home", device, f"{mountpoint}/home"])
    pmb.chroot.root(["mount", "-o", "subvol=@root", device, f"{mountpoint}/root"])
    pmb.chroot.root(["mount", "-o", "subvol=@srv", device, f"{mountpoint}/srv"])
    pmb.chroot.root(["mount", "-o", "subvol=@snapshots", device, f"{mountpoint}/.snapshots"])

    # Disable CoW for /var, to avoid write multiplication
    # and slowdown on databases, containers and VM images.
    pmb.chroot.root(["chattr", "+C", f"{mountpoint}/var"])


def format_and_mount_root(
    device: str, root_label: str, disk: PathString | None, rsync: bool, filesystem: str | None
) -> None:
    """
    :param device: root partition on install block device (e.g. /dev/installp2)
    :param root_label: label of the root partition (e.g. "pmOS_root")
    :param disk: path to disk block device (e.g. /dev/mmcblk0) or None
    """
    # Set default filesystem to ext4
    if filesystem is None:
        filesystem = "ext4"

    is_disk = not rsync and disk is not None
    format_partition_with_filesystem(device, root_label, filesystem, is_disk)

    # Mount
    mountpoint = "/mnt/install"
    logging.info("(native) mount " + device + " to " + mountpoint)
    pmb.chroot.root(["mkdir", "-p", mountpoint])
    pmb.chroot.root(["mount", device, mountpoint])

    if not rsync and filesystem == "btrfs":
        # Make flat btrfs subvolume layout
        prepare_btrfs_subvolumes(device, mountpoint)


def format(
    layout: PartitionLayout | None,
    boot_label: str,
    root_label: str,
    disk: PathString | None,
    rsync: bool,
    filesystem: str,
    full_disk_encryption: bool,
    fde_cipher: str,
    fde_iter_time: str,
) -> None:
    """
    :param layout: partition layout from get_partition_layout() or None
    :param boot_label: label of the boot partition (e.g. "pmOS_boot")
    :param root_label: label of the root partition (e.g. "pmOS_root")
    :param disk: path to disk block device (e.g. /dev/mmcblk0) or None
    """
    if layout:
        uses_prep = layout["prep"] is not None

        if not uses_prep:
            root_dev = f"/dev/installp{layout['root']}"
            boot_dev = f"/dev/installp{layout['boot']}"
        else:
            root_dev = f"/dev/installp{layout['root']}"
            boot_dev = None
    else:
        root_dev = "/dev/install"
        boot_dev = None

    if full_disk_encryption:
        format_luks_root(root_dev, fde_cipher, fde_iter_time)
        root_dev = "/dev/mapper/pm_crypt"

    format_and_mount_root(root_dev, root_label, disk, rsync, filesystem)
    if boot_dev:
        format_and_mount_boot(boot_dev, boot_label)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

import os
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

import pmb.config
import pmb.helpers.pmaports
from pmb.core.pkgrepo import pkgrepo_glob_one, pkgrepo_iglob
from pmb.helpers import logging


def find_path(codename: str, file: str = "") -> Path | None:
    """
    Find path to device APKBUILD under `device/*/device-`.

    :param codename: device codename
    :param file: file to look for (e.g. APKBUILD or deviceinfo), may be empty
    :returns: path to APKBUILD
    """
    g = pkgrepo_glob_one(f"device/*/device-{codename}/{file}")
    if not g:
        return None

    return g


# TODO: This could be simplified using StrEnum once we stop supporting Python 3.10.
class DeviceCategory(Enum):
    """Enum for representing a specific device category."""

    ARCHIVED = "archived"
    DOWNSTREAM = "downstream"
    TESTING = "testing"
    COMMUNITY = "community"
    MAIN = "main"

    @staticmethod
    def shown() -> list[DeviceCategory]:
        """
        Get a list of all device categories that typically are visible, in order of "best" to
        "worst".

        :returns: List of all non-hidden device categories.
        """
        return [
            DeviceCategory.MAIN,
            DeviceCategory.COMMUNITY,
            DeviceCategory.TESTING,
            DeviceCategory.DOWNSTREAM,
        ]

    def allows_downstream_ports(self) -> bool:
        """
        Check whether a given category is allowed to contain downstream ports. This does not
        necessarily mean that it exclusively contains downstream ports.

        :returns: True, if the category allows downstream ports, False if only allows mainline ports.
        """
        match self:
            case DeviceCategory.ARCHIVED | DeviceCategory.DOWNSTREAM:
                return True
            case DeviceCategory.TESTING | DeviceCategory.COMMUNITY | DeviceCategory.MAIN:
                return False
            case _:
                raise AssertionError

    def explain(self) -> str:
        """
        Provide an explanation of a given category.

        :returns: String explaining the given category.
        """
        match self:
            case DeviceCategory.ARCHIVED:
                return "ports that have a better alternative available"
            case DeviceCategory.DOWNSTREAM:
                return "ports that use a downstream kernel — very limited functionality. Not recommended"
            case DeviceCategory.TESTING:
                return 'anything from "just boots in some sense" to almost fully functioning ports'

            case DeviceCategory.COMMUNITY:
                return "often mostly usable, but may lack important functionality"
            case DeviceCategory.MAIN:
                return "ports where mostly everything works"
            case _:
                raise AssertionError

    def color(self) -> str:
        """
        Returns the color associated with the given device category.

        :returns: ANSI escape sequence for the color associated with the given device category.
        """
        styles = pmb.config.styles

        match self:
            case DeviceCategory.ARCHIVED:
                return styles["RED"]
            case DeviceCategory.DOWNSTREAM:
                return styles["YELLOW"]
            case DeviceCategory.TESTING:
                return styles["GREEN"]
            case DeviceCategory.COMMUNITY:
                return styles["BLUE"]
            case DeviceCategory.MAIN:
                return styles["MAGENTA"]
            case _:
                raise AssertionError

    def __str__(self) -> str:
        return self.value


@dataclass(frozen=True)
class DeviceEntry:
    codename: str
    category: DeviceCategory

    def codename_without_vendor(self) -> str:
        return self.codename.split("-", 1)[1]

    def __str__(self) -> str:
        """Remove "vendor-" prefix from device codename and add category."""
        styles = pmb.config.styles
        return f"{self.category.color()}{self.codename_without_vendor()}{styles['END']} ({self.category})"


def list_codenames(vendor: str) -> list[DeviceEntry]:
    """
    Get all devices for which aports are available.

    :param vendor: Vendor name to choose devices from.
    :returns: ["first-device", "second-device", ...]}
    """
    ret: list[DeviceEntry] = []
    for path in pkgrepo_iglob(f"device/*/device-{vendor}-*"):
        codename = os.path.basename(path).split("-", 1)[1]
        # Ensure we don't crash on unknown device categories.
        try:
            category = get_device_category_by_directory_path(path)
        except RuntimeError as exception:
            logging.warning("WARNING: %s: %s", codename, exception)
            continue
        # Get rid of ports inside of hidden device categories.
        if category not in DeviceCategory.shown():
            continue
        ret.append(DeviceEntry(codename, category))
    return ret


def list_vendors() -> set[str]:
    """
    Get all device vendors, for which aports are available.

    :returns: {"vendor1", "vendor2", ...}
    """
    ret = set()
    for path in pkgrepo_iglob("device/*/device-*"):
        vendor = path.name.split("-", 2)[1]
        ret.add(vendor)
    return ret


def get_device_category_by_apkbuild_path(apkbuild_path: Path) -> DeviceCategory:
    """
    Get the category of a device based on the path to its APKBUILD inside of pmaports.

    This will fail to determine the device category from out-of-tree APKBUILDs.

    :apkbuild_path: Path to an APKBUILD within pmaports for a particular device.
    :returns: The device category of the provided device APKBUILD.
    """
    # Path is something like this:
    # .../device/community/device-samsung-m0/APKBUILD
    #            ↑         ↑ parent 1
    #            | parent 2
    category_str = apkbuild_path.parent.parent.name

    try:
        device_category = DeviceCategory(category_str)
    except ValueError as exception:
        raise RuntimeError(f'Unknown device category "{category_str}"') from exception

    return device_category


def get_device_category_by_directory_path(device_directory: Path) -> DeviceCategory:
    """
    Get the category of a device based on the path to its directory inside of pmaports.

    :device_directory: Path to the device package directory for a particular device.
    :returns: The device category of the provided device directory.
    """
    device_apkbuild_path = device_directory / "APKBUILD"

    return get_device_category_by_apkbuild_path(device_apkbuild_path)


def get_device_category_by_name(device_name: str) -> DeviceCategory:
    """
    Get the category of a device based on its name.

    :device_name: Name of a particular device to determine the category of.
                  Format should be "vendor-codename".
    :returns: The device category of the provided device name.
    """
    device_directory = pmb.helpers.pmaports.find(f"device-{device_name}")

    return get_device_category_by_directory_path(device_directory)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
from pathlib import Path

import pmb.build
import pmb.chroot
import pmb.helpers.pmaports
import pmb.helpers.run
from pmb.core import Chroot
from pmb.helpers import logging


def update(pkgname: str) -> None:
    """Fetch all sources and update the checksums in the APKBUILD."""
    pmb.build.init_abuild_minimal()
    pmb.build.copy_to_buildpath(pkgname, no_override=True)
    logging.info("(native) generate checksums for " + pkgname)
    pmb.chroot.user(["abuild", "checksum"], working_dir=Path("/home/pmos/build"))

    # Copy modified APKBUILD back
    source = Chroot.native() / "home/pmos/build/APKBUILD"
    target = f"{os.fspath(pmb.helpers.pmaports.find(pkgname))}/"
    pmb.helpers.run.user(["cp", source, target])


def verify(pkgname: str) -> None:
    """Fetch all sources and verify their checksums."""
    pmb.build.init_abuild_minimal()
    pmb.build.copy_to_buildpath(pkgname, no_override=True)
    logging.info("(native) verify checksums for " + pkgname)

    # Fetch and verify sources, "fetch" alone does not verify them:
    # https://github.com/alpinelinux/abuild/pull/86
    pmb.chroot.user(["abuild", "fetch", "verify"], working_dir=Path("/home/pmos/build"))



# Copyright 2026 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import sys
from pathlib import Path

import pmb.ci
import pmb.helpers.git
from pmb.helpers import logging


def ci(scripts: str, all: bool, fast: bool) -> None:
    topdir = pmb.helpers.git.get_topdir(Path.cwd())
    if not os.path.exists(topdir):
        logging.error(
            "ERROR: change your current directory to a git"
            " repository (e.g. pmbootstrap, pmaports) before running"
            " 'pmbootstrap ci'."
        )
        sys.exit(1)

    scripts_available = pmb.ci.get_ci_scripts(topdir)
    scripts_available = pmb.ci.sort_scripts_by_speed(scripts_available)
    if not scripts_available:
        logging.error(
            "ERROR: no supported CI scripts found in current git"
            " repository, see https://postmarketos.org/pmb-ci"
        )
        sys.exit(1)

    scripts_selected = {}
    if scripts:
        if all:
            raise RuntimeError("Combining --all with script names doesn't make sense")
        for script in scripts:
            if script not in scripts_available:
                logging.error(
                    f"ERROR: script '{script}' not found in git"
                    " repository, found these:"
                    f" {', '.join(scripts_available.keys())}"
                )
                sys.exit(1)
            scripts_selected[script] = scripts_available[script]
    elif all:
        scripts_selected = scripts_available

    if fast:
        for script, script_data in scripts_available.items():
            if "slow" not in script_data["options"]:
                scripts_selected[script] = script_data

    if not pmb.helpers.git.clean_worktree(topdir):
        logging.warning("WARNING: this git repository has uncommitted changes")

    if not scripts_selected:
        scripts_selected = pmb.ci.ask_which_scripts_to_run(scripts_available)

    pmb.ci.run_scripts(topdir, scripts_selected)



# Copyright 2024 Caleb Connolly
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import time

import pmb.helpers.repo
import pmb.parse.apkindex
from pmb.core.arch import Arch
from pmb.helpers import logging

"""Various internal test commands for performance testing and debugging."""


def apkindex_parse_all() -> None:
    indexes = pmb.helpers.repo.apkindex_files(Arch.native())

    pkgs = 0
    indxs = len(indexes)
    start = time.time()
    for index in indexes:
        ret = pmb.parse.apkindex.parse(index)
        pkgs += len(ret)
    end = time.time()
    logging.info(f"Parsed {pkgs} packages from {indxs} APKINDEX files in {end - start:.3f} seconds")


def test(action: str) -> None:
    if action == "apkindex_parse_all":
        apkindex_parse_all()



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import datetime
import enum
import os
import shlex
from pathlib import Path

import pmb.build
import pmb.chroot
import pmb.config.pmaports
import pmb.helpers.pmaports
import pmb.helpers.run
import pmb.parse.apkindex
import pmb.parse.version
from pmb.core import Chroot
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.types import Apkbuild


def copy_to_buildpath(
    package: str, chroot: Chroot = Chroot.native(), no_override: bool = False
) -> None:
    # Sanity check
    aport = pmb.helpers.pmaports.find(package)
    if not os.path.exists(aport / "APKBUILD"):
        raise ValueError(f"Path does not contain an APKBUILD file: {aport}")

    # Clean up folder
    build = chroot / "home/pmos/build"
    if build.exists():
        pmb.helpers.run.root(["rm", "-rf", build])

    # Copy aport contents with resolved symlinks
    pmb.helpers.run.root(["mkdir", "-p", build])
    for entry in aport.iterdir():
        file = entry.name
        # Don't copy those dirs, as those have probably been generated by running `abuild`
        # on the host system directly and not cleaning up after itself.
        # Those dirs might contain broken symlinks and cp fails resolving them.
        if file in ["src", "pkg"]:
            logging.warning(f"WARNING: Not copying {file}, looks like a leftover from abuild")
            continue
        pmb.helpers.run.root(["cp", "-rL", aport / file, build / file])

    if not no_override:
        abuild_overrides(build / "APKBUILD")

    pmb.chroot.root(["chown", "-R", "pmos:pmos", "/home/pmos/build"], chroot)


def abuild_overrides(apkbuild: Path) -> None:
    """Override some abuild functions by patching the APKBUILD file."""
    if apkbuild.is_relative_to(get_context().config.work / "cache_git"):
        raise ValueError(f"Refusing to patch file in pmaports repo: {apkbuild}")

    # Patch the APKBUILD file
    override_path = pmb.config.pmb_src / "pmb/data/abuild_overrides.sh"
    pmb.helpers.run.root(["sh", "-c", f"cat {override_path} >> {apkbuild}"])


class BuildStatus(enum.Enum):
    # The binary package is outdated
    OUTDATED = "outdated"
    # There is no binary package
    NEW = "new"
    # Source package can't be built
    CANT_BUILD = "cant_build"
    # Building isn't needed
    UNNECESSARY = "unnecessary"

    def __str__(self) -> str:
        return self.value

    def necessary(self) -> bool:
        return self in [BuildStatus.OUTDATED, BuildStatus.NEW]


def get_status(arch: Arch | None, apkbuild: Apkbuild) -> BuildStatus:
    """
    Check if the package has already been built.

    Compared to abuild's check, this check also works for different architectures.

    :param arch: package target architecture
    :param apkbuild: from pmb.parse.apkbuild()
    :param indexes: list of APKINDEX.tar.gz paths
    :returns: boolean
    """
    package = apkbuild["pkgname"]
    version_pmaports = apkbuild["pkgver"] + "-r" + apkbuild["pkgrel"]
    msg = "Build is necessary for package '" + package + "': "

    # Get version from APKINDEX
    index_data = pmb.parse.apkindex.package(package, arch, False)
    if not index_data:
        logging.debug(msg + "No binary package available")
        return BuildStatus.NEW

    # Can't build pmaport for arch: use Alpine's package (#1897)
    if arch and not pmb.helpers.pmaports.check_arches(apkbuild["arch"], arch):
        logging.verbose(
            f"{package}: build is not necessary, because pmaport"
            f" can't be built for {arch}. Using Alpine's binary"
            " package."
        )
        return BuildStatus.CANT_BUILD

    # a) Binary repo has a newer version
    version_binary = index_data.version
    if pmb.parse.version.compare(version_binary, version_pmaports) == 1:
        logging.warning(
            f"WARNING: about to install {package} {version_binary}"
            f" (local pmaports: {version_pmaports}, consider"
            " 'pmbootstrap pull')"
        )
        return BuildStatus.UNNECESSARY

    # b) Local pmaports has a newer version
    if version_pmaports != version_binary:
        logging.debug(
            f"{msg}binary package out of date (binary: "
            f"{version_binary}, local pmaports: {version_pmaports})"
        )
        return BuildStatus.OUTDATED

    # Local pmaports and binary repo have the same version
    return BuildStatus.UNNECESSARY


def index_repo(arch: Arch | None = None) -> None:
    """
    Recreate the APKINDEX.tar.gz for a specific repo, and clear the parsing
    cache for that file for the current pmbootstrap session (to prevent
    rebuilding packages twice, in case the rebuild takes less than a second).

    :param arch: when not defined, re-index all repos
    """
    pmb.build.init()

    paths: list[Path] = []

    for channel in pmb.config.pmaports.all_channels():
        pkgdir: Path = get_context().config.work / "packages" / channel
        if arch:
            paths.append(pkgdir / arch)
        else:
            paths += list(pkgdir.glob("*"))

    for path in paths:
        if path.is_dir():
            path_channel, path_arch = path.parts[-2:]
            path_repo_chroot = Path("/mnt/pmbootstrap/packages") / path_channel / path_arch
            logging.debug(f"(native) index {path_channel}/{path_arch} repository")
            description = str(datetime.datetime.now())
            commands = [
                # Wrap the index command with sh so we can use '*.apk'
                [
                    "sh",
                    "-c",
                    "apk -q index --output APKINDEX.tar.gz_"
                    " --description " + shlex.quote(description) + ""
                    " --rewrite-arch " + shlex.quote(path_arch) + " *.apk",
                ],
                ["abuild-sign", "APKINDEX.tar.gz_"],
                ["mv", "APKINDEX.tar.gz_", "APKINDEX.tar.gz"],
            ]
            pmb.chroot.userm(commands, working_dir=path_repo_chroot)
        else:
            logging.debug(f"NOTE: Can't build index for: {path}")
        pmb.parse.apkindex.clear_cache(path / "APKINDEX.tar.gz")


def configure_abuild(chroot: Chroot, verify: bool = False) -> None:
    """
    Set the correct JOBS count in ``abuild.conf``.

    :param verify: internally used to test if changing the config has worked.
    """
    jobs = get_context().config.jobs
    path = chroot / "etc/abuild.conf"
    prefix = "export JOBS="
    with path.open(encoding="utf-8") as handle:
        for line in handle:
            if not line.startswith(prefix):
                continue
            if line != (prefix + str(jobs) + "\n"):
                if verify:
                    raise RuntimeError(
                        f"Failed to configure abuild: {path}"
                        "\nTry to delete the file"
                        "(or zap the chroot)."
                    )
                pmb.chroot.root(
                    ["sed", "-i", "-e", f"s/^{prefix}.*/{prefix}{jobs}/", "/etc/abuild.conf"],
                    chroot,
                )
                configure_abuild(chroot, True)
            return
    pmb.chroot.root(["sed", "-i", f"$ a\\{prefix}{jobs}", "/etc/abuild.conf"], chroot)


def configure_ccache(chroot: Chroot = Chroot.native(), verify: bool = False) -> None:
    """
    Set the maximum ccache size.

    :param verify: internally used to test if changing the config has worked.
    """
    # Check if the settings have been set already
    arch = chroot.arch
    config = get_context().config
    path = config.work / f"cache_ccache_{arch}" / "ccache.conf"
    if os.path.exists(path):
        with open(path, encoding="utf-8") as handle:
            for line in handle:
                if line == ("max_size = " + config.ccache_size + "\n"):
                    return
    if verify:
        raise RuntimeError(
            f"Failed to configure ccache: {path}\nTry to delete the file (or zap the chroot)."
        )

    # Set the size and verify
    pmb.chroot.user(["ccache", "--max-size", config.ccache_size], chroot)
    configure_ccache(chroot, True)



# Copyright 2026 Hugo Posnic
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.helpers.logging


def work_migrate() -> None:
    # do nothing (pmb/__init__.py already did the migration)
    pmb.helpers.logging.disable()



# Copyright 2024 Caleb Connolly
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import pmb.config
from pmb.core.context import get_context
from pmb.helpers import run
from pmb.types import PathString, RunOutputTypeDefault


def log(clear_log: bool, lines: int) -> None:
    context = get_context()
    log_testsuite = pmb.config.pmb_src / ".pytest_tmp/log_testsuite.txt"

    if clear_log:
        run.user(["truncate", "-s", "0", context.log])
        if log_testsuite.exists():
            run.user(["truncate", "-s", "0", log_testsuite])

    cmd: list[PathString] = ["tail", "-n", str(lines), "-F"]

    # Follow the testsuite's log file too if it exists. It will be created when
    # starting a test case that writes to it (git -C test grep log_testsuite).
    if log_testsuite.exists():
        cmd += [log_testsuite]

    # tail writes the last lines of the files to the terminal. Put the regular
    # log at the end, so that output is visible at the bottom (where the user
    # looks for an error / what's currently going on).
    cmd += [context.log]

    run.user(cmd, output=RunOutputTypeDefault.TUI)



# Copyright 2026 Hugo Posnic
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.helpers.repo
from pmb.core.arch import Arch
from pmb.helpers import logging


def update(arch: Arch | None, non_existing: str) -> None:
    existing_only = not non_existing
    if not pmb.helpers.repo.update(arch, True, existing_only):
        logging.info(
            "No APKINDEX files exist, so none have been updated."
            " The pmbootstrap command downloads the APKINDEX files on"
            " demand."
        )
        logging.info(
            "If you want to force downloading the APKINDEX files for"
            " all architectures (not recommended), use:"
            " pmbootstrap update --non-existing"
        )



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import filecmp
import os

import pmb.chroot
import pmb.config
import pmb.config.workdir
import pmb.helpers.apk
import pmb.helpers.apk_static
import pmb.helpers.repo
import pmb.helpers.run
from pmb.core import Chroot, ChrootType
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.meta import Cache
from pmb.types import PathString


def copy_resolv_conf(chroot: Chroot) -> None:
    """
    Use pythons super fast file compare function (due to caching)
    and copy the /etc/resolv.conf to the chroot, in case it is
    different from the host.
    If the file doesn't exist, create an empty file with 'touch'.
    """
    host = "/etc/resolv.conf"
    resolv_path = chroot / host
    if os.path.exists(host):
        if not resolv_path.exists() or not filecmp.cmp(host, resolv_path):
            pmb.helpers.run.root(["cp", host, resolv_path])
    else:
        pmb.helpers.run.root(["touch", resolv_path])


def mark_in_chroot(chroot: Chroot = Chroot.native()) -> None:
    """
    Touch a flag so we can know when we're running in chroot (and
    don't accidentally flash partitions on our host). This marker
    gets removed in pmb.chroot.shutdown (pmbootstrap shutdown).
    """
    in_chroot_file = chroot / "in-pmbootstrap"
    if not in_chroot_file.exists():
        pmb.helpers.run.root(["touch", in_chroot_file])


def init_keys() -> None:
    """
    All Alpine and postmarketOS repository keys are shipped with pmbootstrap.
    Copy them into $WORK/config_apk_keys, which gets mounted inside the various
    chroots as /etc/apk/keys.

    This is done before installing any package, so apk can verify APKINDEX
    files of binary repositories even though alpine-keys/postmarketos-keys are
    not installed yet.
    """
    for key in pmb.config.apk_keys_path.glob("*.pub"):
        target = get_context().config.work / "config_apk_keys" / key.name
        if not target.exists():
            # Copy as root, so the resulting files in chroots are owned by root
            pmb.helpers.run.root(["cp", key, target])


@Cache()
def warn_if_chroots_outdated() -> None:
    outdated = pmb.config.workdir.chroots_outdated()
    if outdated:
        days_warn = int(pmb.config.chroot_outdated / 3600 / 24)
        msg = ""
        if Chroot.native() in outdated:
            msg += "your native"
            if Chroot.rootfs(get_context().config.device) in outdated:
                msg += " and rootfs chroots are"
            else:
                msg += " chroot is"
        elif Chroot.rootfs(get_context().config.device) in outdated:
            msg += "your rootfs chroot is"
        else:
            msg += "some of your chroots are"
        logging.warning(
            f"WARNING: {msg} older than {days_warn} days. Consider running 'pmbootstrap zap'."
        )


@Cache("chroot")
def init(chroot: Chroot) -> None:
    """
    Initialize a chroot by copying the resolv.conf and updating
    /etc/apk/repositories. If /bin/sh is missing, create the chroot from
    scratch.
    """
    # When already initialized: just prepare the chroot
    arch = chroot.arch

    # If the channel is wrong and the user has auto_zap_misconfigured_chroots
    # enabled, zap the chroot and reinitialize it
    if chroot.exists():
        zap = pmb.config.workdir.chroot_check_channel(chroot)
        if zap:
            pmb.chroot.del_chroot(chroot.path, confirm=False)
            pmb.config.workdir.clean()

    pmb.chroot.mount(chroot)
    mark_in_chroot(chroot)
    if chroot.exists():
        copy_resolv_conf(chroot)
        pmb.helpers.apk.update_repository_list(chroot.path)
        warn_if_chroots_outdated()
        return

    # Fetch apk.static
    pmb.helpers.apk_static.init()

    logging.info(f"({chroot}) Creating chroot")

    # Initialize /etc/apk/keys/, resolv.conf, repositories
    init_keys()
    copy_resolv_conf(chroot)
    pmb.helpers.apk.update_repository_list(chroot.path)

    pmb.config.workdir.chroot_save_init(chroot)

    pmb.helpers.repo.update(arch)
    # Create the /usr-merge-related symlinks, which needs to be done manually
    if pmb.config.pmaports.read_config().get("supported_usr_merge", False):
        pmb.helpers.run.root(
            [
                "mkdir",
                "-p",
                f"{chroot.path}/usr/bin",
                f"{chroot.path}/usr/sbin",
                f"{chroot.path}/usr/lib",
            ]
        )
        pmb.helpers.run.root(["ln", "-s", "usr/bin", "usr/sbin", "usr/lib", f"{chroot.path}/"])
    # Create the bin-merge-related symlinks, which are done manually to be
    # consistent with the /usr-merge
    if pmb.config.pmaports.read_config().get("supported_bin_merge", False):
        pmb.helpers.run.root(
            [
                "rm",
                "-r",
                f"{chroot.path}/sbin",
                f"{chroot.path}/usr/sbin",
            ]
        )
        pmb.helpers.run.root(["ln", "-s", "usr/bin", f"{chroot.path}/sbin"])
        pmb.helpers.run.root(["ln", "-s", "bin", f"{chroot.path}/usr/sbin"])
    # Install minimal amount of things to get a functional chroot.
    # We don't use alpine-base since it depends on openrc, and neither
    # postmarketos-base, since that's quite big (e.g: contains an init system)
    pkgs = ["alpine-baselayout", "apk-tools", "busybox", "musl-utils"]
    cmd: list[PathString] = ["--initdb"]
    pmb.helpers.apk.run([*cmd, "add", *pkgs], chroot)

    # Building chroots: create "pmos" user, add symlinks to /home/pmos
    if chroot.type != ChrootType.ROOTFS:
        pmb.chroot.root(["adduser", "-D", "pmos", "-u", pmb.config.chroot_uid_user], chroot)

        # Create the links (with subfolders if necessary)
        for target, link_name in pmb.config.chroot_home_symlinks.items():
            link_dir = os.path.dirname(link_name)
            if not os.path.exists(chroot / link_dir):
                pmb.chroot.user(["mkdir", "-p", link_dir], chroot)
            if not os.path.exists(chroot / target):
                pmb.chroot.root(["mkdir", "-p", target], chroot)
            pmb.chroot.user(["ln", "-s", target, link_name], chroot)
            pmb.chroot.root(["chown", "pmos:pmos", target], chroot)
